/* Plastik widget style for KDE 3
   Copyright (C) 2003 Sandro Giessl <ceebx@users.sourceforge.net>

   based on the KDE style "dotNET":

   Copyright (C) 2001-2002, Chris Lee <clee@kde.org>
                            Carsten Pfeiffer <pfeiffer@kde.org>
                            Karol Szwed <gallium@kde.org>
   Drawing routines completely reimplemented from KDE3 HighColor, which was
   originally based on some stuff from the KDE2 HighColor.

   based on drawing routines of the style "Keramik":

   Copyright (c) 2002 Malte Starostik <malte@kde.org>
             (c) 2002,2003 Maksim Orlovich <mo002j@mail.rochester.edu>
   based on the KDE3 HighColor Style
   Copyright (C) 2001-2002 Karol Szwed      <gallium@kde.org>
             (C) 2001-2002 Fredrik Hglund  <fredrik@kde.org>
   Drawing routines adapted from the KDE2 HCStyle,
   Copyright (C) 2000 Daniel M. Duley       <mosfet@kde.org>
             (C) 2000 Dirk Mueller          <mueller@kde.org>
             (C) 2001 Martijn Klingens      <klingens@kde.org>
   Progressbar code based on KStyle,
   Copyright (C) 2001-2002 Karol Szwed <gallium@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License version 2 as published by the Free Software Foundation.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
 */

#include <qstylefactory.h>
#include <qpointarray.h>
#include <qpainter.h>
#include <qtabbar.h>
#include <qprogressbar.h>
#include <qcombobox.h>
#include <qlistbox.h>
#include <qscrollbar.h>
#include <qstyleplugin.h>
#include <qpushbutton.h>
#include <qtoolbutton.h>
#include <qtoolbar.h>
#include <qmenubar.h>
#include <qpopupmenu.h>
#include <qdrawutil.h>
#include <qapplication.h>
#include <qvariant.h>
#include <qpixmapcache.h>
#include <qregion.h>
#include <qslider.h>
#include <qsettings.h>
#include <kpixmap.h>
#include <kpixmapeffect.h>

#include "plastik.h"
#include "plastik.moc"
#include "misc.h"

// some bitmaps for the radio button so it's easier to handle the circle stuff...
// 13x13
static unsigned char radiobutton_mask_bits[] = {
   0xf0, 0x01, 0xfc, 0x07, 0xfe, 0x0f, 0xfe, 0x0f, 0xff, 0x1f, 0xff, 0x1f,
   0xff, 0x1f, 0xff, 0x1f, 0xff, 0x1f, 0xfe, 0x0f, 0xfe, 0x0f, 0xfc, 0x07,
   0xf0, 0x01};
static unsigned char radiobutton_contour_bits[] = {
   0xe0, 0x00, 0x18, 0x03, 0x04, 0x04, 0x02, 0x08, 0x02, 0x08, 0x01, 0x10,
   0x01, 0x10, 0x01, 0x10, 0x02, 0x08, 0x02, 0x08, 0x04, 0x04, 0x18, 0x03,
   0xe0, 0x00};
static unsigned char radiobutton_aa_inside_bits[] = {
   0x00, 0x00, 0xa0, 0x00, 0x08, 0x02, 0x04, 0x04, 0x00, 0x00, 0x02, 0x08,
   0x00, 0x00, 0x02, 0x08, 0x00, 0x00, 0x04, 0x04, 0x08, 0x02, 0xa0, 0x00,
   0x00, 0x00};
static unsigned char radiobutton_aa_outside_bits[] = {
   0x10, 0x01, 0x04, 0x04, 0x02, 0x08, 0x00, 0x00, 0x01, 0x10, 0x00, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x01, 0x10, 0x00, 0x00, 0x02, 0x08, 0x04, 0x04,
   0x10, 0x01};
static unsigned char radiobutton_highlight1_bits[] = {
   0x00, 0x00, 0xe0, 0x00, 0x18, 0x03, 0x04, 0x04, 0x04, 0x04, 0x02, 0x08,
   0x02, 0x08, 0x02, 0x08, 0x04, 0x04, 0x04, 0x04, 0x18, 0x03, 0xe0, 0x00,
   0x00, 0x00};
static unsigned char radiobutton_highlight2_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0xe0, 0x00, 0x18, 0x03, 0x08, 0x02, 0x04, 0x04,
   0x04, 0x04, 0x04, 0x04, 0x08, 0x02, 0x18, 0x03, 0xe0, 0x00, 0x00, 0x00,
   0x00, 0x00};
// check mark
const uint CHECKMARKSIZE = 9; // 9x9
static unsigned char checkmark_aa_bits[] = {
   0x45, 0x01, 0x28, 0x00, 0x11, 0x01, 0x82, 0x00, 0x44, 0x00, 0x82, 0x00,
   0x11, 0x01, 0x28, 0x00, 0x45, 0x01};
static unsigned char checkmark_dark_bits[] = {
   0x82, 0x00, 0x45, 0x01, 0xaa, 0x00, 0x54, 0x00, 0x28, 0x00, 0x74, 0x00,
   0xea, 0x00, 0xc5, 0x01, 0x82, 0x00};
static unsigned char checkmark_light_bits[] = {
   0x00, 0xfe, 0x82, 0xfe, 0x44, 0xfe, 0x28, 0xfe, 0x10, 0xfe, 0x08, 0xfe,
   0x04, 0xfe, 0x02, 0xfe, 0x00, 0xfe};
// radio mark
const uint RADIOMARKSIZE = 9; // 9x9
static unsigned char radiomark_aa_bits[] = {
   0x00, 0x00, 0x44, 0x00, 0x82, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
   0x82, 0x00, 0x44, 0x00, 0x00, 0x00};
static unsigned char radiomark_dark_bits[] = {
   0x00, 0x00, 0x38, 0x00, 0x44, 0x00, 0xf2, 0x00, 0xfa, 0x00, 0xfa, 0x00,
   0x7c, 0x00, 0x38, 0x00, 0x00, 0x00};
static unsigned char radiomark_light_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x0c, 0x00, 0x04, 0x00, 0x04, 0x00,
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

// -- Style Plugin Interface -------------------------
class PlastikStylePlugin : public QStylePlugin
{
    public:
        PlastikStylePlugin() {}
        ~PlastikStylePlugin() {}

        QStringList keys() const {
            return QStringList() << "Plastik";
        }

        QStyle* create( const QString& key ) {
            if (key == "plastik")
                return new PlastikStyle;
            return 0;
        }
};

Q_EXPORT_PLUGIN( PlastikStylePlugin )
// -- end --

PlastikStyle::PlastikStyle() : KStyle( AllowMenuTransparency, ThreeButtonScrollBar)
{
    hoverWidget = 0;
    kickerMode = false;
    kornMode = false;

    reverseLayout = QApplication::reverseLayout();

//     if (qApp->inherits("KApplication")) {
//         connect( qApp, SIGNAL( kdisplayPaletteChanged() ), SLOT( paletteChanged() ));
//     }
}


PlastikStyle::~PlastikStyle()
{
}

bool PlastikStyle::inheritsKHTML(const QWidget* w) const
{
    if (w->parentWidget(true) && w->parentWidget(true)->parentWidget(true) && w->parentWidget(true)->parentWidget(true)->parentWidget(true) &&  w->parentWidget(true)->parentWidget(true)->parentWidget(true)->inherits("KHTMLView")) {
        return true;
    } else {
        return false;
    }
}

void PlastikStyle::polish(QApplication* app)
{
    if (!qstrcmp(app->argv()[0], "kicker"))
        kickerMode = true;
    else if (!qstrcmp(app->argv()[0], "korn"))
        kornMode = true;
}

void PlastikStyle::polish(QWidget* widget)
{
    if (widget->inherits("QPushButton") || widget->inherits("QComboBox") ||
            widget->inherits("QSpinWidget") || widget->inherits("QSlider") ||
            widget->inherits("QCheckBox") || widget->inherits("QRadioButton") )
    {
        widget->installEventFilter(this);
    }

    KStyle::polish(widget);
}

void PlastikStyle::unPolish(QWidget* widget)
{
    if (widget->inherits("QPushButton") || widget->inherits("QComboBox") ||
            widget->inherits("QSpinWidget") || widget->inherits("QSlider") ||
            widget->inherits("QCheckBox") || widget->inherits("QRadioButton") )
    {
        widget->removeEventFilter(this);
    }

    KStyle::unPolish(widget);
}

void PlastikStyle::renderContour(QPainter *p,
                                const QRect &r,
                                const QColor &backgroundColor,
                                const QColor &contour,
                                const uint flags) const
{
    const bool drawLeft = flags&Draw_Left;
    const bool drawRight = flags&Draw_Right;
    const bool drawTop = flags&Draw_Top;
    const bool drawBottom = flags&Draw_Bottom;
    const bool disabled = flags&Is_Disabled;

    QColor contourColor;
    if (disabled) {
        contourColor = backgroundColor.dark(150);
    } else {
        contourColor = contour;
    }
    p->save();

// sides
    p->setPen( alphaBlendColors(backgroundColor, contourColor, 50) );
    if(drawLeft)
        p->drawLine(r.left(), drawTop?r.top()+2:r.top(), r.left(), drawBottom?r.bottom()-2:r.bottom());
    if(drawRight)
        p->drawLine(r.right(), drawTop?r.top()+2:r.top(), r.right(), drawBottom?r.bottom()-2:r.bottom());
    if(drawTop)
        p->drawLine(drawLeft?r.left()+2:r.left(), r.top(), drawRight?r.right()-2:r.right(), r.top());
    if(drawBottom)
        p->drawLine(drawLeft?r.left()+2:r.left(), r.bottom(), drawRight?r.right()-2:r.right(), r.bottom());

// edges
    if(drawLeft && drawTop) {
        switch(flags&Round_UpperLeft) {
            case false:
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 50) );
                p->drawPoint(r.left()+1, r.top());
                p->drawPoint(r.left(), r.top()+1);
                // anti-alias
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 150) );
                p->drawPoint(r.left(), r.top());
                break;
            case true:
            default:
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 50) );
                p->drawPoint(r.left()+1, r.top()+1);
                // anti-alias
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 150) );
                p->drawPoint(r.left()+1, r.top());
                p->drawPoint(r.left(), r.top()+1);
                // TODO: is this possible to do with a mask?
                // last free point: background
                p->setPen(backgroundColor);
                p->drawPoint(r.left(), r.top());
                break;
        }
    }
    if(drawLeft && drawBottom) {
        switch(flags&Round_BottomLeft) {
            case false:
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 50) );
                p->drawPoint(r.left()+1, r.bottom());
                p->drawPoint(r.left(), r.bottom()-1);
                // anti-alias
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 150) );
                p->drawPoint(r.left(), r.bottom());
                break;
            case true:
            default:
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 50) );
                p->drawPoint(r.left()+1, r.bottom()-1);
                // anti-alias
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 150) );
                p->drawPoint(r.left()+1, r.bottom());
                p->drawPoint(r.left(), r.bottom()-1);
                // last free point: background
                p->setPen(backgroundColor);
                p->drawPoint(r.left(), r.bottom());
                break;
        }
    }
    if(drawRight && drawTop) {
        switch(flags&Round_UpperRight) {
            case false:
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 50) );
                p->drawPoint(r.right()-1, r.top());
                p->drawPoint(r.right(), r.top()+1);
                // anti-alias
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 150) );
                p->drawPoint(r.right(), r.top());
                break;
            case true:
            default:
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 50) );
                p->drawPoint(r.right()-1, r.top()+1);
                // anti-alias
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 150) );
                p->drawPoint(r.right()-1, r.top());
                p->drawPoint(r.right(), r.top()+1);
                // last free point: background
                p->setPen(backgroundColor);
                p->drawPoint(r.right(), r.top());
                break;
        }
    }
    if(drawRight && drawBottom) {
        switch(flags&Round_BottomRight) {
            case false:
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 50) );
                p->drawPoint(r.right()-1, r.bottom());
                p->drawPoint(r.right(), r.bottom()-1);
                // anti-alias
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 150) );
                p->drawPoint(r.right(), r.bottom());
                break;
            case true:
            default:
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 50) );
                p->drawPoint(r.right()-1, r.bottom()-1);
                // anti-alias
                p->setPen(alphaBlendColors(backgroundColor, contourColor, 150) );
                p->drawPoint(r.right()-1, r.bottom());
                p->drawPoint(r.right(), r.bottom()-1);
                // last free point: background
                p->setPen(backgroundColor);
                p->drawPoint(r.right(), r.bottom());
                break;
        }
    }

    p->restore();
}

void PlastikStyle::renderMask(QPainter *p,
                              const QRect &r,
                              const QColor &color,
                              const uint flags) const
{
    const bool roundUpperLeft = flags&Round_UpperLeft;
    const bool roundUpperRight = flags&Round_UpperRight;
    const bool roundBottomLeft = flags&Round_BottomLeft;
    const bool roundBottomRight = flags&Round_BottomRight;

    p->save();

    p->fillRect (QRect(r.x()+1, r.y()+1, r.width()-2, r.height()-2) , color);

    p->setPen(color);
    // sides
    p->drawLine(roundUpperLeft?r.x()+1:r.x(), r.y(),
                roundUpperRight?r.right()-1:r.right(), r.y() );
    p->drawLine(roundBottomLeft?r.x()+1:r.x(), r.bottom(),
                roundBottomRight?r.right()-1:r.right(), r.bottom() );
    p->drawLine(r.x(), roundUpperLeft?r.y()+1:r.y(),
                r.x(), roundBottomLeft?r.bottom()-1:r.bottom() );
    p->drawLine(r.right(), roundUpperLeft?r.y()+1:r.y(),
                r.right(), roundBottomLeft?r.bottom()-1:r.bottom() );

    p->restore();
}

void PlastikStyle::renderSourface(QPainter *p,
                                 const QRect &r,
                                 const QColor &backgroundColor,
                                 const QColor &buttonColor,
                                 const QColor &highlightColor,
                                 int intensity,
                                 const uint flags) const
{
    const bool drawLeft = flags&Draw_Left;
    const bool drawRight = flags&Draw_Right;
    const bool drawTop = flags&Draw_Top;
    const bool drawBottom = flags&Draw_Bottom;
    const bool roundUpperLeft = flags&Round_UpperLeft;
    const bool roundUpperRight = flags&Round_UpperRight;
    const bool roundBottomLeft = flags&Round_BottomLeft;
    const bool roundBottomRight = flags&Round_BottomRight;
    const bool sunken = flags&Is_Sunken;
    const bool horizontal = flags&Is_Horizontal;
    const bool disabled = flags&Is_Disabled;
    bool highlight = false,
         highlightLeft = false,
         highlightRight = false,
         highlightTop = false,
         highlightBottom = false;
    // only highlight if not sunken & not disabled...
    if(!sunken && !disabled) {
        highlight = (flags&Is_Highlight);
        highlightLeft = (flags&Highlight_Left);
        highlightRight = (flags&Highlight_Right);
        highlightTop = (flags&Highlight_Top);
        highlightBottom = (flags&Highlight_Bottom);
    }

    KPixmap tempPixmap;

    p->save();

    QColor baseColor = alphaBlendColors(backgroundColor, disabled?backgroundColor:buttonColor, 10);
    if (disabled) {
        intensity = 2;
    } else if (highlight) {
        // blend this _slightly_ with the background
        baseColor = alphaBlendColors(baseColor, highlightColor, 240);
    } else if (sunken) {
        // enforce a common sunken-style...
        baseColor = baseColor.dark(100+intensity);
        intensity = 2;
    }
// some often needed colors...
    // 1 more intensive than 2 and 3.
    const QColor colorTop1 = alphaBlendColors(baseColor,
                       sunken?baseColor.dark(100+intensity*2):baseColor.light(100+intensity*2), 80);
    const QColor colorTop2 = alphaBlendColors(baseColor,
                       sunken?baseColor.dark(100+intensity):baseColor.light(100+intensity), 80);
    const QColor colorBottom1 = alphaBlendColors(baseColor,
                          sunken?baseColor.light(100+intensity*2):baseColor.dark(100+intensity*2), 80);
    const QColor colorBottom2 = alphaBlendColors(baseColor,
                          sunken?baseColor.light(100+intensity):baseColor.dark(100+intensity), 80);

// sides
    if (drawLeft) {
        if (horizontal) {
            int height = r.height();
            if (roundUpperLeft || !drawTop) height--;
            if (roundBottomLeft || !drawBottom) height--;
            tempPixmap.resize(1, height);
            KPixmapEffect::gradient(tempPixmap,
                                    colorTop1, baseColor,
                                    KPixmapEffect::VerticalGradient);
            p->drawPixmap(r.left(), (roundUpperLeft&&drawTop)?r.top()+1:r.top(), tempPixmap);
        } else {
            p->setPen(colorTop1 );
            p->drawLine(r.left(), (roundUpperLeft&&drawTop)?r.top()+1:r.top(),
                        r.left(), (roundBottomLeft&&drawBottom)?r.bottom()-1:r.bottom() );
        }
    }
    if (drawRight) {
        if (horizontal) {
            int height = r.height();
            // TODO: there's still a bogus in it: when edge4 is Thick
            //       and we don't whant to draw the Top, we have a unpainted area
            if (roundUpperRight || !drawTop) height--;
            if (roundBottomRight || !drawBottom) height--;
            tempPixmap.resize(1, height);
            KPixmapEffect::gradient(tempPixmap,
                                    baseColor, colorBottom1,
                                    KPixmapEffect::VerticalGradient);
            p->drawPixmap(r.right(), (roundUpperRight&&drawTop)?r.top()+1:r.top(), tempPixmap);
        } else {
            p->setPen(colorBottom1 );
            p->drawLine(r.right(), (roundUpperRight&&drawTop)?r.top()+1:r.top(),
                        r.right(), (roundBottomRight&&drawBottom)?r.bottom()-1:r.bottom() );
        }
    }
    if (drawTop) {
        if (horizontal) {
            p->setPen(colorTop1 );
            p->drawLine((roundUpperLeft&&drawLeft)?r.left()+1:r.left(), r.top(),
                        (roundUpperRight&&drawRight)?r.right()-1:r.right(), r.top() );
        } else {
            int width = r.width();
            if (roundUpperLeft || !drawLeft) width--;
            if (roundUpperRight || !drawRight) width--;
            tempPixmap.resize(width, 1);
            KPixmapEffect::gradient(tempPixmap,
                                    colorTop1, colorTop2,
                                    KPixmapEffect::HorizontalGradient);
            p->drawPixmap((roundUpperLeft&&drawLeft)?r.left()+1:r.left(), r.top(), tempPixmap);
        }
    }
    if (drawBottom) {
        if (horizontal) {
            p->setPen(colorBottom1 );
            p->drawLine((roundBottomLeft&&drawLeft)?r.left()+1:r.left(), r.bottom(),
                        (roundBottomRight&&drawRight)?r.right()-1:r.right(), r.bottom() );
        } else {
            int width = r.width();
            if (roundBottomLeft || !drawLeft) width--;
            if (roundBottomRight || !drawRight) width--;
            tempPixmap.resize(width, 1);
            KPixmapEffect::gradient(tempPixmap,
                                    colorBottom2, colorBottom1,
                                    KPixmapEffect::HorizontalGradient);
            p->drawPixmap((roundBottomLeft&&drawLeft)?r.left()+1:r.left(), r.bottom(), tempPixmap);
        }
    }

// button area...
    int width = r.width();
    int height = r.height();
    if (drawLeft) width--;
    if (drawRight) width--;
    if (drawTop) height--;
    if (drawBottom) height--;
    tempPixmap.resize(width, height);
    if (horizontal) {
        KPixmapEffect::gradient(tempPixmap,
                                colorTop2, colorBottom2,
                                KPixmapEffect::VerticalGradient);
    } else {
        KPixmapEffect::gradient(tempPixmap,
                                colorTop2, colorBottom2,
                                KPixmapEffect::HorizontalGradient);
    }
    p->drawPixmap(drawLeft?r.left()+1:r.left(), drawTop?r.top()+1:r.top(), tempPixmap);

// highlighting...
    if(highlightTop) {
        p->setPen(alphaBlendColors(colorTop1 , highlightColor, 80) );
        p->drawLine((roundUpperLeft&&drawLeft)?r.left()+1:r.left(), r.top(),
                    (roundUpperRight&&drawRight)?r.right()-1:r.right(), r.top() );
        p->setPen(alphaBlendColors(colorTop2 , highlightColor, 150) );
        p->drawLine(highlightLeft?r.left()+1:r.left(), r.top()+1,
                    highlightRight?r.right()-1:r.right(), r.top()+1 );
    }
    if(highlightBottom) {
        p->setPen(alphaBlendColors(colorBottom1 , highlightColor, 80) );
        p->drawLine((roundBottomLeft&&drawLeft)?r.left()+1:r.left(), r.bottom(),
                    (roundBottomRight&&drawRight)?r.right()-1:r.right(), r.bottom() );
        p->setPen(alphaBlendColors(colorBottom2 , highlightColor, 150) );
        p->drawLine(highlightLeft?r.left()+1:r.left(), r.bottom()-1,
                    highlightRight?r.right()-1:r.right(), r.bottom()-1 );
    }
    if(highlightLeft) {
        p->setPen(alphaBlendColors(colorTop1 , highlightColor, 80) );
        p->drawLine(r.left(), (roundUpperLeft&&drawTop)?r.top()+1:r.top(),
                    r.left(), (roundBottomLeft&&drawBottom)?r.bottom()-1:r.bottom() );
        p->setPen(alphaBlendColors(colorTop2 , highlightColor, 150) );
        p->drawLine(r.left()+1, highlightTop?r.top()+1:r.top(),
                    r.left()+1, highlightBottom?r.bottom()-1:r.bottom() );
    }
    if(highlightRight) {
        p->setPen(alphaBlendColors(colorBottom1 , highlightColor, 80) );
        p->drawLine(r.right(), (roundUpperRight&&drawTop)?r.top()+1:r.top(),
                    r.right(), (roundBottomRight&&drawBottom)?r.bottom()-1:r.bottom() );
        p->setPen(alphaBlendColors(colorBottom2 , highlightColor, 150) );
        p->drawLine(r.right()-1, highlightTop?r.top()+1:r.top(),
                    r.right()-1, highlightBottom?r.bottom()-1:r.bottom() );
    }

    p->restore();
}

void PlastikStyle::renderButton(QPainter *p,
                               const QRect &r,
                               const QColorGroup &g,
                               bool sunken,
                               bool mouseOver,
                               bool horizontal,
                               bool enabled) const
{
    // small fix for the kicker buttons...
    if(kickerMode) enabled = true;

    p->save();

    uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|
            Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight;
    if(!enabled) contourFlags|=Is_Disabled;
    renderContour(p, r, g.background(), g.button().dark(150),
                  contourFlags);

    uint sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|
                         Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight;
    if(horizontal) sourfaceFlags|=Is_Horizontal;
    if(!enabled) sourfaceFlags|=Is_Disabled;
    else {
        if(sunken) sourfaceFlags|=Is_Sunken;
        else {
            if(mouseOver) {
                sourfaceFlags|=Is_Highlight;
                if(horizontal) {
                    sourfaceFlags|=Highlight_Top;
                    sourfaceFlags|=Highlight_Bottom;
                } else {
                    sourfaceFlags|=Highlight_Left;
                    sourfaceFlags|=Highlight_Right;
                }
            }
        }
    }
    renderSourface(p, QRect(r.left()+1, r.top()+1, r.width()-2, r.height()-2),
                   g.background(), g.button(), g.highlight(), 8, sourfaceFlags);

    p->restore();
}

void PlastikStyle::renderHeader(QPainter *p,
                                const QRect &r,
                                const QColorGroup &g,
                                bool sunken,
                                bool mouseOver,
                                bool horizontal,
                                bool enabled) const
{
    if(kickerMode) enabled = true;

    p->save();

    uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
    // Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight;
    if(!enabled) contourFlags|=Is_Disabled;
    renderContour(p, r, g.background(), g.button().dark(150),
                    contourFlags);

    uint sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
    // Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight;
    if(horizontal) sourfaceFlags|=Is_Horizontal;
    if(!enabled) sourfaceFlags|=Is_Disabled;
    else {
        if(sunken) sourfaceFlags|=Is_Sunken;
        else {
            if(mouseOver) {
                sourfaceFlags|=Is_Highlight;
                if(horizontal) {
                    sourfaceFlags|=Highlight_Top;
                    sourfaceFlags|=Highlight_Bottom;
                } else {
                    sourfaceFlags|=Highlight_Left;
                    sourfaceFlags|=Highlight_Right;
                }
            }
        }
    }
    renderSourface(p, QRect(r.left()+1, r.top()+1, r.width()-2, r.height()-2),
                    g.background(), g.button(), g.highlight(), 8,
                    sourfaceFlags);

    p->restore();

}

void PlastikStyle::renderSliderHandle(QPainter *p,
                                     const QRect &r,
                                     const QColorGroup &g,
                                     const bool mouseOver,
                                     const bool horizontal,
                                     const bool enabled) const
{
    const QColor contourColor = g.button().dark(150);
    const QColor contourColor3 = g.background().dark(150);
    KPixmap tempPixmap;

    int xcenter = (r.left()+r.right()) / 2;
    int ycenter = (r.top()+r.bottom()) / 2;

    p->save();

    if (horizontal) {
        renderContour(p, QRect(xcenter-5, ycenter-6, 11, 10),
                      g.background(), enabled?contourColor:contourColor3,
                      Draw_Left|Draw_Right|Draw_Top|Round_UpperLeft|Round_UpperRight);

        // manual contour: vertex
        p->setPen(alphaBlendColors(g.background(), enabled?contourColor:contourColor3, 50) );
        p->drawPoint(xcenter-5+1, ycenter+4);
        p->drawPoint(xcenter+5-1, ycenter+4);
        p->drawPoint(xcenter-5+2, ycenter+5);
        p->drawPoint(xcenter+5-2, ycenter+5);
        p->drawPoint(xcenter-5+3, ycenter+6);
        p->drawPoint(xcenter+5-3, ycenter+6);
        p->drawPoint(xcenter-5+4, ycenter+7);
        p->drawPoint(xcenter+5-4, ycenter+7);
        // anti-aliasing of the contour... sort of. :)
        p->setPen(alphaBlendColors(g.background(), enabled?contourColor:contourColor3, 100) );
        p->drawPoint(xcenter, ycenter+8);
        p->setPen(alphaBlendColors(g.background(), enabled?contourColor:contourColor3, 150) );
        p->drawPoint(xcenter-5, ycenter+4);
        p->drawPoint(xcenter+5, ycenter+4);
        p->drawPoint(xcenter-5+1, ycenter+5);
        p->drawPoint(xcenter+5-1, ycenter+5);
        p->drawPoint(xcenter-5+2, ycenter+6);
        p->drawPoint(xcenter+5-2, ycenter+6);
        p->drawPoint(xcenter-5+3, ycenter+7);
        p->drawPoint(xcenter+5-3, ycenter+7);


        QRegion mask(xcenter-4, ycenter-5, 9, 13);
        mask -= QRegion(xcenter-4, ycenter+4, 1, 4);
        mask -= QRegion(xcenter-3, ycenter+5, 1, 3);
        mask -= QRegion(xcenter-2, ycenter+6, 1, 2);
        mask -= QRegion(xcenter-1, ycenter+7, 1, 1);
        mask -= QRegion(xcenter+1, ycenter+7, 1, 1);
        mask -= QRegion(xcenter+2, ycenter+6, 1, 2);
        mask -= QRegion(xcenter+3, ycenter+5, 1, 3);
        mask -= QRegion(xcenter+4, ycenter+4, 1, 4);
        p->setClipRegion(mask);
        p->setClipping(true);
        uint sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Round_UpperLeft|Round_UpperRight|Is_Horizontal;
        if(mouseOver) {
            sourfaceFlags |= Is_Highlight;
            sourfaceFlags |= Highlight_Top;
        }
        renderSourface(p, QRect(xcenter-4, ycenter-5, 9, 13),
                       g.background(), enabled?g.button():g.background(), g.highlight(),
                       enabled?10:0, sourfaceFlags);
        p->setClipping(false);
    } else {
        renderContour(p, QRect(xcenter-6, ycenter-5, 10, 11),
                      g.background(), enabled?contourColor:contourColor3,
                      Draw_Left|Draw_Top|Draw_Bottom|Round_UpperLeft|Round_BottomLeft);

        // manual contour: vertex
        p->setPen(alphaBlendColors(g.background(), enabled?contourColor:contourColor3, 50) );
        p->drawPoint(xcenter+4, ycenter-5+1);
        p->drawPoint(xcenter+4, ycenter+5-1);
        p->drawPoint(xcenter+5, ycenter-5+2);
        p->drawPoint(xcenter+5, ycenter+5-2);
        p->drawPoint(xcenter+6, ycenter-5+3);
        p->drawPoint(xcenter+6, ycenter+5-3);
        p->drawPoint(xcenter+7, ycenter-5+4);
        p->drawPoint(xcenter+7, ycenter+5-4);
        // anti-aliasing. ...sort of :)
        p->setPen(alphaBlendColors(g.background(), enabled?contourColor:contourColor3, 100) );
        p->drawPoint(xcenter+8, ycenter);
        p->setPen(alphaBlendColors(g.background(), enabled?contourColor:contourColor3, 150) );
        p->drawPoint(xcenter+4, ycenter-5);
        p->drawPoint(xcenter+4, ycenter+5);
        p->drawPoint(xcenter+5, ycenter-5+1);
        p->drawPoint(xcenter+5, ycenter+5-1);
        p->drawPoint(xcenter+6, ycenter-5+2);
        p->drawPoint(xcenter+6, ycenter+5-2);
        p->drawPoint(xcenter+7, ycenter-5+3);
        p->drawPoint(xcenter+7, ycenter+5-3);

        QRegion mask(xcenter-5, ycenter-4, 13, 9);
        mask -= QRegion(xcenter+4, ycenter-4, 4, 1);
        mask -= QRegion(xcenter+5, ycenter-3, 3, 1);
        mask -= QRegion(xcenter+6, ycenter-2, 2, 1);
        mask -= QRegion(xcenter+7, ycenter-1, 1, 1);
        mask -= QRegion(xcenter+7, ycenter+1, 1, 1);
        mask -= QRegion(xcenter+6, ycenter+2, 2, 1);
        mask -= QRegion(xcenter+5, ycenter+3, 3, 1);
        mask -= QRegion(xcenter+4, ycenter+4, 4, 1);
        p->setClipRegion(mask);
        p->setClipping(true);
        uint sourfaceFlags = Draw_Left|Draw_Top|Draw_Bottom|Round_UpperLeft|Round_BottomLeft|
                           Round_UpperRight|Is_Horizontal;
        if(mouseOver) {
            sourfaceFlags |= Is_Highlight;
            sourfaceFlags |= Highlight_Left;
        }
        renderSourface(p, QRect(xcenter-5, ycenter-4, 13, 9),
                       g.background(), enabled?g.button():g.background(), g.highlight(),
                       enabled?10:0, sourfaceFlags);
        p->setClipping(false);
    }

    p->restore();
}

void PlastikStyle::renderDot(QPainter *p,
                             const QPoint &point,
                             const QColor &baseColor,
                             const bool thick,
                             const bool sunken) const
{
    const QColor topColor = alphaBlendColors(baseColor, sunken?baseColor.dark(130):baseColor.light(150), 70);
    const QColor bottomColor = alphaBlendColors(baseColor, sunken?baseColor.light(150):baseColor.dark(130), 70);
    p->setPen(topColor );
    p->drawLine(point.x(), point.y(), point.x()+1, point.y());
    p->drawPoint(point.x(), point.y()+1);
    p->setPen(bottomColor );
    if(thick) {
        p->drawLine(point.x()+1, point.y()+2, point.x()+2, point.y()+2);
        p->drawPoint(point.x()+2, point.y()+1);
    } else {
        p->drawPoint(point.x()+1, point.y()+1);
    }
}

void PlastikStyle::renderPanel(QPainter *p,
                              const QRect &r,
                              const QColorGroup &g,
                              const bool pseudo3d,
                              const bool sunken,
                              const bool thick) const
{
    int x, x2, y, y2, w, h;
    r.rect(&x,&y,&w,&h);
    r.coords(&x, &y, &x2, &y2);

    p->save();

    renderContour(p, r, g.background(), g.background().dark(180));

    if(pseudo3d) {
        if (sunken) {
            p->setPen(alphaBlendColors(g.background(), g.background().dark(160), 110) );
        } else {
            p->setPen(alphaBlendColors(g.background(), g.background().light(160), 110) );
        }
        p->drawLine(r.left()+2, r.top()+1, r.right()-2, r.top()+1);
        p->drawLine(r.left()+1, r.top()+2, r.left()+1, r.bottom()-2);
        if (sunken) {
            p->setPen(alphaBlendColors(g.background(), g.background().light(160), 110) );
        } else {
            p->setPen(alphaBlendColors(g.background(), g.background().dark(160), 110) );
        }
        p->drawLine(r.left()+2, r.bottom()-1, r.right()-2, r.bottom()-1);
        p->drawLine(r.right()-1, r.top()+2, r.right()-1, r.bottom()-2);
    }

    if (kickerMode) {
    //  Stolen wholesale from Keramik. I don't like it, but oh well.
    // TODO: not exactly sure what this does.
        if (!thick) {
            if (sunken) {
                const QCOORD corners[] = { x2, y, x2, y2, x, y2, x, y };
                p->setPen(g.background().dark());
                p->drawConvexPolygon(QPointArray(4, corners));
                p->setPen(g.background().light());
                p->drawPolyline(QPointArray(4, corners), 0, 3);
            } else {
                const QCOORD corners[] = { x, y2, x, y, x2, y, x2, y2 };
                p->setPen(g.background().dark());
                p->drawPolygon(QPointArray(4, corners));
                p->setPen(g.background().light());
                p->drawPolyline(QPointArray(4, corners), 0, 3);
            }
        }
    }

    p->restore();
}

void PlastikStyle::renderTab(QPainter *p,
                            const QRect &r,
                            const QColorGroup &g,
                            bool mouseOver,
                            const bool selected,
                            const bool bottom,
                            const TabPosition pos) const
{
    QColor contourColor = g.button().dark(150);
    QColor contourColor2 = g.background().dark(180); // should be the same as used in renderPanel
    KPixmap tempPixmap;

    // wasn't possible for me to track the mouse with eventFilter() to highlight at the
    // right time, so this is disabled for now. :-(
    mouseOver = false;

    p->save();

    if(!bottom) {
        if (selected) {
        // is selected
            renderContour(p, QRect(r.left()+1, r.top(), r.width()-2, r.height()-3),
                          g.background(), contourColor2,
                          Draw_Left|Draw_Right|Draw_Top|Round_UpperLeft|Round_UpperRight|
                          Round_BottomLeft|Round_BottomRight);

            // some manual contour below the tab
            p->setPen( alphaBlendColors(g.background(), contourColor2, 50) );
            p->drawPoint(r.left(), r.bottom()-2);
            p->drawPoint(r.right(), r.bottom()-2);
            // somewhat anti-aliasing for the manual contour...
            p->setPen( alphaBlendColors(g.background(), contourColor2, 150) );
            p->drawPoint(r.left()+1, r.bottom()-2);
            if (pos == First || pos == Single) {
                p->drawPoint(r.left(), r.bottom()-3);
            } else {
                p->drawPoint(r.left(), r.bottom()-1);
            }
            p->drawPoint(r.right()-1, r.bottom()-2);
            p->drawPoint(r.right(), r.bottom()-1);

            if (pos == First || pos == Single) {
                // fake parts of the panel border...
                // contour
                p->setPen(alphaBlendColors(g.background(), contourColor2, 50) );
                p->drawLine(r.left(), r.bottom()-1, r.left(), r.bottom());
                // light...
                p->setPen(alphaBlendColors(g.background(), g.background().light(160), 110) );
                p->drawLine(r.left()+1, r.bottom()-1, r.left()+1, r.bottom());
            }

            // sourface
            p->setPen(alphaBlendColors(g.background(), g.background().light(160), 100) );
            p->drawLine(r.left()+3, r.top()+1, r.right()-3, r.top()+1 );
            tempPixmap.resize(1,r.height()-6);
            KPixmapEffect::gradient(tempPixmap,
                                    alphaBlendColors(g.background(), g.background().light(160), 110),
                                    alphaBlendColors(g.background(), g.background().light(150), 110),
                                    KPixmapEffect::VerticalGradient);
            p->drawPixmap(r.left()+2, r.top()+2, tempPixmap);

            tempPixmap.resize(1,r.height()-6);
            KPixmapEffect::gradient(tempPixmap,
                                    alphaBlendColors(g.background(), g.background().dark(160), 110),
                                    alphaBlendColors(g.background(), g.background().dark(150), 110),
                                    KPixmapEffect::VerticalGradient);
            p->drawPixmap(r.right()-2, r.top()+2, tempPixmap);

            // TODO: see if it is better to draw a gradient or not
            //       IMHO it's not, since the thing basically is part of the panel...

            // draw parts of the inactive tabs around...
            if(pos != Single) {
                p->setPen(alphaBlendColors(g.background(), contourColor, 50) );
                if(pos != First)
                    p->drawPoint(r.left(), r.top()+3);
                if(pos != Last)
                    p->drawPoint(r.right(), r.top()+3);
                if(pos != First) {
                    renderSourface(p, QRect(r.left(), r.top()+4, 1, r.height()-7),
                                g.background(), g.button(), g.highlight(), 5,
                                Draw_Top|Draw_Bottom|Is_Horizontal);
                }
                if(pos != Last) {
                    renderSourface(p, QRect(r.right(), r.top()+4, 1, r.height()-7),
                                g.background(), g.button(), g.highlight(), 5,
                                Draw_Top|Draw_Bottom|Is_Horizontal);
                }
            }

        } else if (pos == First) {
        // is the first inactive tab
            renderContour(p, QRect(r.left(), r.top()+3, r.width(), r.height()-5),
                          g.background(), contourColor,
                          Draw_Left|Draw_Right|Draw_Top|Round_UpperLeft);

            // complete contour manual
            p->setPen(alphaBlendColors(g.background(), contourColor, 50) );
            p->drawLine(r.left(), r.bottom()-2, r.left(), r.bottom());

            // "fake" parts of the panel border...
            // contour
            p->setPen(alphaBlendColors(g.background(), contourColor2, 50) );
            p->drawLine(r.left()+2, r.bottom()-1, r.right(), r.bottom()-1);
            p->drawPoint(r.left()+1, r.bottom());
            // anti-aliasing of the contour...
            p->setPen(alphaBlendColors(g.background(), contourColor2, 150) );
            p->drawPoint(r.left()+1, r.bottom()-1);
            // light...
            p->setPen(alphaBlendColors(g.background(), g.background().light(160), 110) );
            p->drawLine(r.left()+2, r.bottom(), r.right(), r.bottom());

            uint sourfaceFlags = Draw_Left|Draw_Top|Draw_Bottom|Round_UpperLeft|Is_Horizontal;
            if(mouseOver) {
                sourfaceFlags |= Highlight_Top;
                sourfaceFlags |= Is_Highlight;
            }
            renderSourface(p, QRect(r.left()+1, r.top()+4, r.width()-2, r.height()-6),
                           g.background(), g.button(), g.highlight(), 5,
                           sourfaceFlags);
        } else if (pos == Last) {
        // is the last inactive tab
            renderContour(p, QRect(r.left()+1, r.top()+3, r.width()-1, r.height()-5),
                          g.background(), contourColor,
                          Draw_Right|Draw_Top|Round_UpperRight);

            // fake parts of the panel border
            p->setPen(alphaBlendColors(g.background(), contourColor2, 50) );
            p->drawLine(r.left(), r.bottom()-1, r.right(), r.bottom()-1);
            p->setPen(alphaBlendColors(g.background(), g.background().light(160), 110) );
            p->drawLine(r.left(), r.bottom(), r.right(), r.bottom());

            uint sourfaceFlags = Draw_Right|Draw_Top|Draw_Bottom|Round_UpperRight|Is_Horizontal;
            if(mouseOver) {
                sourfaceFlags |= Highlight_Top;
                sourfaceFlags |= Is_Highlight;
            }
            renderSourface(p, QRect(r.left()+1, r.top()+4, r.width()-2, r.height()-6),
                           g.background(), g.button(), g.highlight(), 5,
                           sourfaceFlags);
        } else {
        // is a middle inactive tab
            renderContour(p, QRect(r.left()+1, r.top()+3, r.width()-1, r.height()-5),
                          g.background(), contourColor,
                          Draw_Right|Draw_Top);

            // "fake" parts of the panel border...
            p->setPen(alphaBlendColors(g.background(), contourColor2, 50) );
            p->drawLine(r.left()+1, r.bottom()-1, r.right(), r.bottom()-1);
            p->setPen(alphaBlendColors(g.background(), g.background().light(160), 110) );
            p->drawLine(r.left()+1, r.bottom(), r.right(), r.bottom());

            uint sourfaceFlags = Draw_Top|Draw_Bottom|Is_Horizontal;
            if(mouseOver) {
                sourfaceFlags |= Highlight_Top;
                sourfaceFlags |= Is_Highlight;
            }
            renderSourface(p, QRect(r.left()+1, r.top()+4, r.width()-2, r.height()-6),
                           g.background(), g.button(), g.highlight(), 5,
                           sourfaceFlags);
        }

    } else {
    // Tabs are at the bottom

        if (selected) {
        // is selected
            renderContour(p, QRect(r.left()+1, r.top()+3, r.width()-2, r.height()-3),
                          g.background(), contourColor2,
                          Draw_Left|Draw_Right|Draw_Bottom|Round_BottomLeft|Round_BottomRight);

            // some manual contour below the tab
            p->setPen( alphaBlendColors(g.background(), contourColor2, 50) );
            p->drawPoint(r.left(), r.top()+2);
            p->drawPoint(r.right(), r.top()+2);
            // somewhat anti-aliasing for the manual contour...
            p->setPen( alphaBlendColors(g.background(), contourColor2, 150) );
            p->drawPoint(r.left()+1, r.top()+2);
            if (pos != First) {
                p->drawPoint(r.left(), r.top()+1);
            } else {
                p->drawPoint(r.left(), r.top()+3);
            }
            p->drawPoint(r.right()-1, r.top()+2);
            p->drawPoint(r.right(), r.top()+1);

            if (pos == First) {
                // fake parts of the panel border...
                // contour
                p->setPen(alphaBlendColors(g.background(), contourColor2, 50) );
                p->drawLine(r.left(), r.top()+1, r.left(), r.top());
                // light...
                p->setPen(alphaBlendColors(g.background(), g.background().light(160), 110) );
                p->drawLine(r.left()+1, r.top()+1, r.left()+1, r.top());
            }

            // sourface
            p->setPen(alphaBlendColors(g.background(), g.background().dark(160), 100) );
            p->drawLine(r.left()+3, r.bottom()-1, r.right()-3, r.bottom()-1 );
            tempPixmap.resize(1,r.height()-6);
            KPixmapEffect::gradient(tempPixmap,
                                    alphaBlendColors(g.background(), g.background().light(160), 110),
                                    alphaBlendColors(g.background(), g.background().light(150), 110),
                                    KPixmapEffect::VerticalGradient);
            p->drawPixmap(r.left()+2, r.top()+4, tempPixmap);

            tempPixmap.resize(1,r.height()-6);
            KPixmapEffect::gradient(tempPixmap,
                                    alphaBlendColors(g.background(), g.background().dark(160), 110),
                                    alphaBlendColors(g.background(), g.background().dark(150), 110),
                                    KPixmapEffect::VerticalGradient);
            p->drawPixmap(r.right()-2, r.top()+4, tempPixmap);

            // draw parts of the inactive tabs around...
            p->setPen(alphaBlendColors(g.background(), contourColor, 50) );
            if(pos != First)
                p->drawPoint(r.left(), r.bottom()-3);
            if(pos != Last)
                p->drawPoint(r.right(), r.bottom()-3);
            if(pos != First) {
                renderSourface(p, QRect(r.left(), r.top()+3, 1, r.height()-7),
                               g.background(), g.button(), g.highlight(), 5,
                               Draw_Top|Is_Horizontal);
            }
            if(pos != Last) {
                renderSourface(p, QRect(r.right(), r.top()+3, 1, r.height()-7),
                               g.background(), g.button(), g.highlight(), 5,
                               Draw_Top|Is_Horizontal);
            }

        } else if (pos == First) {
        // is the first inactive tab
            renderContour(p, QRect(r.left(), r.top()+2, r.width(), r.height()-5),
                          g.background(), contourColor,
                          Draw_Left|Draw_Right|Draw_Bottom|Round_BottomLeft);

            // complete contour manual
            p->setPen(alphaBlendColors(g.background(), contourColor, 50) );
            p->drawLine(r.left(), r.top()+2, r.left(), r.top());

            // "fake" parts of the panel border...
            // contour
            p->setPen(alphaBlendColors(g.background(), contourColor2, 50) );
            p->drawLine(r.left()+2, r.top()+1, r.right(), r.top()+1);
            p->drawPoint(r.left()+1, r.top());
            // anti-aliasing of the contour...
            p->setPen(alphaBlendColors(g.background(), contourColor2, 150) );
            p->drawPoint(r.left()+1, r.top()+1);
            // light...
            p->setPen(alphaBlendColors(g.background(), g.background().dark(160), 110) );
            p->drawLine(r.left()+2, r.top(), r.right(), r.top());

            uint sourfaceFlags = Draw_Left|Draw_Top|Draw_Bottom|Round_BottomLeft|Is_Horizontal;
            if(mouseOver) {
                sourfaceFlags |= Highlight_Bottom;
                sourfaceFlags |= Is_Highlight;
            }
            renderSourface(p, QRect(r.left()+1, r.top()+2, r.width()-2, r.height()-6),
                           g.background(), g.button(), g.highlight(), 5,
                           sourfaceFlags);
        } else if (pos == Last) {
        // is the last inactive tab
            renderContour(p, QRect(r.left()+1, r.top()+2, r.width()-1, r.height()-5),
                          g.background(), contourColor,
                          Draw_Right|Draw_Bottom|Round_BottomRight);

            // fake parts of the panel border
            p->setPen(alphaBlendColors(g.background(), contourColor2, 50) );
            p->drawLine(r.left(), r.top()+1, r.right(), r.top()+1);
            p->setPen(alphaBlendColors(g.background(), g.background().dark(160), 110) );
            p->drawLine(r.left(), r.top(), r.right(), r.top());

            uint sourfaceFlags = Draw_Right|Draw_Top|Draw_Bottom|Round_BottomRight|Is_Horizontal;
            if(mouseOver) {
                sourfaceFlags |= Highlight_Bottom;
                sourfaceFlags |= Is_Highlight;
            }
            renderSourface(p, QRect(r.left()+1, r.top()+2, r.width()-2, r.height()-6),
                           g.background(), g.button(), g.highlight(), 5,
                           sourfaceFlags);
        } else {
        // is a middle inactive tab
            renderContour(p, QRect(r.left()+1, r.top()+2, r.width()-1, r.height()-5),
                          g.background(), contourColor,
                          Draw_Right|Draw_Bottom);

            // "fake" parts of the panel border...
            p->setPen(alphaBlendColors(g.background(), contourColor2, 50) );
            p->drawLine(r.left()+1, r.top()+1, r.right(), r.top()+1);
            p->setPen(alphaBlendColors(g.background(), g.background().dark(160), 110) );
            p->drawLine(r.left()+1, r.top(), r.right(), r.top());

            uint sourfaceFlags = Draw_Top|Draw_Bottom|Is_Horizontal;
            if(mouseOver) {
                sourfaceFlags |= Highlight_Bottom;
                sourfaceFlags |= Is_Highlight;
            }
            renderSourface(p, QRect(r.left()+1, r.top()+2, r.width()-2, r.height()-6),
                           g.background(), g.button(), g.highlight(), 5,
                           sourfaceFlags);
        }
    }

    p->restore();
}



void PlastikStyle::drawKStylePrimitive(KStylePrimitive kpe,
                                      QPainter *p,
                                      const QWidget* widget,
                                      const QRect &r,
                                      const QColorGroup &cg,
                                      SFlags flags,
                                      const QStyleOption& opt) const
{
    // some "global" vars...
    const bool enabled = (flags & Style_Enabled);

//  SLIDER
//  ------
    switch( kpe ) {
        case KPE_SliderGroove: {
            const QSlider* slider = (const QSlider*)widget;
            bool horizontal = slider->orientation() == Horizontal;

            if (horizontal) {
                int center = r.height()/2;
                renderContour(p, QRect(r.left(), center-2, r.width(), 4),
                              cg.background(), cg.background().dark(enabled?150:130),
                              Draw_Left|Draw_Right|Draw_Top|Draw_Bottom);
            } else {
                int center = r.width()/2;
                renderContour(p, QRect(center-2, r.top(), 4, r.height()),
                              cg.background(), cg.background().dark(enabled?150:130),
                              Draw_Left|Draw_Right|Draw_Top|Draw_Bottom);
            }
            break;
        }

        case KPE_SliderHandle: {
                const QSlider* slider = (const QSlider*)widget;
                bool horizontal = slider->orientation() == Horizontal;
                renderSliderHandle(p, r, cg, (hoverWidget==widget), horizontal, enabled);
                break;
            }

        default:
            KStyle::drawKStylePrimitive(kpe, p, widget, r, cg, flags, opt);
    }
}


void PlastikStyle::drawPrimitive(PrimitiveElement pe,
                                QPainter *p,
                                const QRect &r,
                                const QColorGroup &cg,
                                SFlags flags,
                                const QStyleOption &opt ) const
{
    bool down   = flags & Style_Down;
    bool on     = flags & Style_On;
    bool sunken = flags & Style_Sunken;
    bool horiz  = flags & Style_Horizontal;
    const bool enabled = flags & Style_Enabled;
    const bool mouseOver = flags & Style_MouseOver;

    int x = r.x();
    int y = r.y();
    int w = r.width();
    int h = r.height();

    int x2, y2;
    r.coords(&x, &y, &x2, &y2);

    switch(pe) {
    // BUTTONS
    // -------
        case PE_FocusRect: {
            p->drawWinFocusRect( r );
            break;
        }

        case PE_HeaderSection: {
            // the taskbar buttons seems to be painted with PE_HeaderSection but I
            // want them look like normal buttons (at least for now. :) )
            if(!kickerMode) {
                renderHeader(p, r, cg, (on||down), mouseOver, true, enabled );
                break;
            }
        }
        case PE_ButtonBevel:
        case PE_ButtonTool:
        case PE_ButtonDropDown:
        case PE_ButtonCommand: {
            renderButton(p, r, cg, (on||down), mouseOver, true, enabled );
            break;
        }

        case PE_ButtonDefault: {
            // empty for now.
            break;
        }

        case PE_ScrollBarSlider: {
            uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            if(!enabled) contourFlags|=Is_Disabled;
            renderContour(p, r, cg.background(), cg.button().dark(150),
                    contourFlags);

            uint sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            if(horiz) sourfaceFlags|=Is_Horizontal;
// // doesn't look very well....
//             if(!enabled) sourfaceFlags|=Is_Disabled;
            renderSourface(p, QRect(r.left()+1, r.top()+1, r.width()-2, r.height()-2),
                        cg.background(), cg.button(), cg.background(),
                        enabled?10:5, sourfaceFlags);

            // deco TODO: imho this still doesn't look very well
            if(!horiz) {
                int n = (r.height()-4-4)/3;
                if(n>5) n=5;
                if(n%2 == 0) n--;
                int ycenter = r.top()+r.height()/2-1;
                for(int j = ycenter-3*((n-1)/2); j <= ycenter+3*((n-1)/2); j+=3) {
                    for(int k = 4; k <= 12; k+=3) {
                        renderDot(p, QPoint(k, j), cg.button() );
                    }
                }
            } else {
                int n = (r.width()-4-4)/3;
                if(n>5) n=5;
                if(n%2 == 0) n--;
                int xcenter = r.left()+r.width()/2-1;
                for(int j = xcenter-3*((n-1)/2); j <= xcenter+3*((n-1)/2); j+=3) {
                    for(int k = 4; k <= 12; k+=3) {
                        renderDot(p, QPoint(j, k), cg.button() );
                    }
                }
            }

            break;
        }

        case PE_ScrollBarAddPage:
        case PE_ScrollBarSubPage: {
            // TODO: imho this is somewhat strange... have a look at this later...
            if (on || down) {
                p->fillRect(r, QBrush(cg.mid().dark()));
            } else {
                p->save();

                if(flags & Style_Horizontal) {
                    p->setPen(cg.background().dark(106));
                    p->drawLine(r.left(), r.top(), r.right(), r.top());
                    p->setPen(cg.background().light(106));
                    p->drawLine(r.left(), r.bottom(), r.right(), r.bottom());
                    p->fillRect(r.left(), r.top()+1, r.width(), r.height()-2,cg.background());
                } else {
                    p->setPen(cg.background().dark(106));
                    p->drawLine(r.left(), r.top(), r.left(), r.bottom());
                    p->setPen(cg.background().light(106));
                    p->drawLine(r.right(), r.top(), r.right(), r.bottom());
                    p->fillRect(r.left()+1, r.top(), r.width()-2, r.height(),cg.background());
                }

                p->restore();
            }
            p->fillRect(r, QBrush(cg.background().light(), Dense4Pattern));
            break;
        }

    // SCROLLBAR BUTTONS
    // -----------------
        case PE_ScrollBarSubLine: {
            uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            uint sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            if(down) sourfaceFlags|=Is_Sunken;
//             if(!enabled) {
//                 contourFlags|=Is_Disabled;
//                 sourfaceFlags|=Is_Disabled;
//             }
            if(horiz) {
                contourFlags |= Round_UpperLeft|Round_BottomLeft|Is_Horizontal;
                sourfaceFlags |= Round_UpperLeft|Round_BottomLeft|Is_Horizontal;
            } else {
                contourFlags |= Round_UpperLeft|Round_UpperRight;
                sourfaceFlags |= Round_UpperLeft|Round_UpperRight;
            }
            renderContour(p, r, cg.background(), cg.button().dark(150),
                    contourFlags);
            renderSourface(p, QRect(r.left()+1, r.top()+1, r.width()-2, r.height()-2),
                    cg.background(), cg.button(), cg.highlight(), enabled?10:5,
                    sourfaceFlags);

            p->setPen(down ? cg.highlightedText() : cg.foreground());
            drawPrimitive((horiz ? PE_ArrowLeft : PE_ArrowUp), p, r, cg, flags);
            break;
        }

        case PE_ScrollBarAddLine: {
            uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            uint sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            if(down) sourfaceFlags|=Is_Sunken;
//             if(!enabled) {
//                 contourFlags|=Is_Disabled;
//                 sourfaceFlags|=Is_Disabled;
//             }
            if(horiz) {
                contourFlags |= Round_UpperRight|Round_BottomRight|Is_Horizontal;
                sourfaceFlags |= Round_UpperRight|Round_BottomRight|Is_Horizontal;
            } else {
                contourFlags |= Round_BottomLeft|Round_BottomRight;
                sourfaceFlags |= Round_BottomLeft|Round_BottomRight;
            }
            renderContour(p, r, cg.background(), cg.button().dark(150),
                    contourFlags);
            renderSourface(p, QRect(r.left()+1, r.top()+1, r.width()-2, r.height()-2),
                    cg.background(), cg.button(), cg.highlight(), enabled?10:5,
                    sourfaceFlags);

            p->setPen(down ? cg.highlightedText() : cg.foreground());
            drawPrimitive((horiz ? PE_ArrowRight : PE_ArrowDown), p, r, cg, flags);
            break;
        }

    // CHECKBOXES
    // ----------
        case PE_Indicator: {
            QColor contentColor = enabled?cg.base():cg.background();

            p->save();

            uint contourFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom;
            if(!enabled) {
                contourFlags |= Is_Disabled;
            }
            renderContour(p, r, cg.background(), cg.button().dark(150),
                    contourFlags);

            // surface
            uint sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Is_Horizontal;
            if(!enabled) {
                sourfaceFlags |= Is_Disabled;
            } else if(mouseOver) {
                contentColor = alphaBlendColors(contentColor, cg.highlight(), 240);
                sourfaceFlags |= Is_Highlight;
                sourfaceFlags |= Highlight_Left|Highlight_Right|
                        Highlight_Top|Highlight_Bottom;
            }
            renderSourface(p, QRect(r.x()+1, r.y()+1, r.width()-2, r.height()-2),
                    cg.background(), contentColor, cg.highlight(), enabled?10:5, sourfaceFlags);

            if (flags & Style_On || flags & Style_Down) {
                drawPrimitive(PE_CheckMark, p, r, cg, flags);
            }

            p->restore();

            break;
        }

        case PE_IndicatorMask: {
            p->fillRect (r, color1);
            break;
        }

    // RADIOBUTTONS
    // ------------
        case PE_ExclusiveIndicator: {
            const QColor contourColor = enabled?cg.button().dark(150):cg.background().dark(150);
            QColor contentColor = enabled?cg.base():cg.background();

            QBitmap bmp;
            bmp = QBitmap(13, 13, radiobutton_mask_bits, true);
            // first the sourface...
            uint sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Is_Horizontal;
            if(!enabled) {
                sourfaceFlags |= Is_Disabled;
            } else if (mouseOver) {
                contentColor = alphaBlendColors(contentColor, cg.highlight(), 240);
            }
            p->setClipRegion(bmp);
            p->setClipping(true);
            renderSourface(p, r,
                    cg.background(), contentColor, cg.highlight(), enabled?10:5, sourfaceFlags);
            p->setClipping(false);

            // ...then contour, anti-alias, mouseOver...
            // contour
            bmp = QBitmap(13, 13, radiobutton_contour_bits, true);
            bmp.setMask(bmp);
            p->setPen(alphaBlendColors(cg.background(), contourColor, 50) );
            p->drawPixmap(x, y, bmp);
            // anti-alias outside
            bmp = QBitmap(13, 13, radiobutton_aa_outside_bits, true);
            bmp.setMask(bmp);
            p->setPen(alphaBlendColors(cg.background(), contourColor, 150) );
            p->drawPixmap(x, y, bmp);
            // highlighting...
            if(mouseOver) {
                bmp = QBitmap(13, 13, radiobutton_highlight1_bits, true);
                bmp.setMask(bmp);
                p->setPen(alphaBlendColors(contentColor, cg.highlight(), 80) );
                p->drawPixmap(x, y, bmp);
                bmp = QBitmap(13, 13, radiobutton_highlight2_bits, true);
                bmp.setMask(bmp);
                p->setPen(alphaBlendColors(contentColor, cg.highlight(), 150) );
                p->drawPixmap(x, y, bmp);
            }
            // anti-alias inside, "above" the higlighting!
            bmp = QBitmap(13, 13, radiobutton_aa_inside_bits, true);
            bmp.setMask(bmp);
            if(mouseOver) {
                p->setPen(alphaBlendColors(cg.highlight(), contourColor, 180) );
            } else {
                p->setPen(alphaBlendColors(contentColor, contourColor, 180) );
            }
            p->drawPixmap(x, y, bmp);

            break;
        }

        case PE_ExclusiveIndicatorMask: {
            p->fillRect(r, color0);

            QBitmap bmp;
            bmp = QBitmap(13, 13, radiobutton_mask_bits, true);
            bmp.setMask(bmp);
            p->setPen(color1);
            p->drawPixmap(x, y, bmp);

            break;
        }

    // GENERAL PANELS
    // --------------
        case PE_Splitter: {
            p->fillRect(r, cg.background());
            p->setBrush(cg.background().dark());

            renderContour(p, r, cg.background(), cg.background().dark(130),
                          Draw_Left|Draw_Right|Draw_Top|Draw_Bottom);
            if (w > h) {
                if (h > 4) {
                    for(int k = 2*r.width()/10; k < 8*r.width()/10; k+=5) {
                        int ycenter = r.height()/2;
                        p->setPen(alphaBlendColors(cg.background(), cg.background().light(160), 110) );
                        p->drawLine(k, ycenter-1, k+1, ycenter-1);
                        p->drawPoint(k, ycenter-1);
                        p->setPen(alphaBlendColors(cg.background(), cg.background().dark(160), 110) );
                        p->drawPoint(k+1, ycenter);
                    }
                }
            } else {
                if (w > 4) {
                    for(int k = 2*r.height()/10; k < 8*r.height()/10; k+=5) {
                        int xcenter = r.width()/2;
                        p->setPen(alphaBlendColors(cg.background(), cg.background().light(150), 70) );
                        p->drawLine(xcenter-1, k, xcenter-1, k+1);
                        p->drawPoint(xcenter, k);
                        p->setPen(alphaBlendColors(cg.background(), cg.background().dark(150), 70) );
                        p->drawPoint(xcenter, k+1);
                    }
                }
            }

                break;
        }

        case PE_PanelGroupBox:
        case PE_GroupBoxFrame: {
                renderPanel(p, r, cg, false);

            break;
        }

        case PE_WindowFrame:
        case PE_Panel: {
            renderPanel(p, r, cg, true, sunken);
            break;
        }

        case PE_PanelLineEdit: {
            renderContour(p, r, cg.background(),
                          enabled?cg.button().dark(150):cg.background().dark(150) );
            const QColor contentColor = enabled?cg.base():cg.background();
            p->setPen(contentColor.dark(130) );
            p->drawLine(r.left()+1, r.top()+2, r.left()+1, r.bottom()-2 );
            p->drawLine(r.left()+2, r.top()+1, r.right()-2, r.top()+1 );
            p->setPen(contentColor.light(130) );
            p->drawLine(r.left()+2, r.bottom()-1, r.right()-2, r.bottom()-1 );
            p->drawLine(r.right()-1, r.top()+2, r.right()-1, r.bottom()-2 );
            break;
        }

        case PE_StatusBarSection: {
            renderContour(p, r, cg.background(), cg.background().dark(160),
                          Draw_Left|Draw_Right|Draw_Top|Draw_Bottom);
            break;
        }

        case PE_TabBarBase: // Still not sure what this one does
        case PE_PanelTabWidget: {
            renderPanel(p, r, cg, true, sunken, true);
            break;
        }

        case PE_PanelPopup: {
            renderContour(p, r, cg.background(), cg.background().dark(180),
                          Draw_Left|Draw_Right|Draw_Top|Draw_Bottom);
            break;
        }

    // MENU / TOOLBAR PANEL
    // --------------------
        case PE_PanelMenuBar:
        case PE_PanelDockWindow: {
            p->fillRect(r, cg.background() );

            break;
        }

    // TOOLBAR/DOCK WINDOW HANDLE
    // --------------------------
        case PE_DockWindowResizeHandle: {
            renderButton(p, r, cg);
            break;
        }

        case PE_DockWindowHandle: {

            p->fillRect(r, cg.background());

            int counter = 1;

            if(horiz) {
                int center = r.left()+r.width()/2;
                for(int j = r.top()+2; j <= r.bottom()-3; j+=3) {
                    if(counter%2 == 0) {
                        renderDot(p, QPoint(center+1, j), cg.background(), true, true);
                    } else {
                        renderDot(p, QPoint(center-2, j), cg.background(), true, true);
                    }
                    counter++;
                }
            } else {
                int center = r.top()+r.height()/2;
                for(int j = r.left()+2; j <= r.right()-3; j+=3) {
                    if(counter%2 == 0) {
                        renderDot(p, QPoint(j, center+1), cg.background(), true, true);
                    } else {
                        renderDot(p, QPoint(j, center-2), cg.background(), true, true);
                    }
                    counter++;
                }
            }

            break;
        }

    // TOOLBAR SEPARATOR
    // -----------------
        case PE_DockWindowSeparator: {
            p->fillRect(r, cg.background());

            if(horiz) {
                int center = r.left()+r.width()/2;
                for(int k = r.top()+1; k <= r.bottom()-2; k+=3)
                {
                    renderDot(p, QPoint(center-1, k), cg.background(), false, true );
                }
            } else {
                int center = r.top()+r.height()/2;
                for(int k = r.left()+1; k <= r.right()-2; k+=3)
                {
                    renderDot(p, QPoint(k, center-1), cg.background(), false, true );
                }
            }
            break;
        }

        case PE_CheckMark: {
            const QColor contentColor = enabled?cg.base():cg.background();
            QColor checkmarkColor = enabled?cg.highlight():cg.background();
            if(flags & Style_Down) {
                checkmarkColor = alphaBlendColors(contentColor, checkmarkColor, 150);
            }

            int x = r.center().x() - 4, y = r.center().y() - 4;
            QBitmap bmp;
            bmp = QBitmap(CHECKMARKSIZE, CHECKMARKSIZE, checkmark_dark_bits, true);
            bmp.setMask(bmp);
            p->setPen(alphaBlendColors(contentColor, checkmarkColor.dark(150), 50) );
            p->drawPixmap(x, y, bmp);
            bmp = QBitmap(CHECKMARKSIZE, CHECKMARKSIZE, checkmark_light_bits, true);
            bmp.setMask(bmp);
            p->setPen(alphaBlendColors(contentColor, checkmarkColor.dark(125), 50) );
            p->drawPixmap(x, y, bmp);
            bmp = QBitmap(CHECKMARKSIZE, CHECKMARKSIZE, checkmark_aa_bits, true);
            bmp.setMask(bmp);
            p->setPen(alphaBlendColors(contentColor, checkmarkColor.dark(150), 150) );
            p->drawPixmap(x, y, bmp);

            break;
        }

        case PE_SpinWidgetUp:
        case PE_SpinWidgetDown:
        case PE_HeaderArrow:
        case PE_ArrowUp:
        case PE_ArrowDown:
        case PE_ArrowLeft:
        case PE_ArrowRight: {
            QPointArray a;

            switch (pe) {
                case PE_SpinWidgetUp:
                case PE_ArrowUp: {
                    a.setPoints(7, u_arrow);
                    break;
                }
                case PE_SpinWidgetDown:
                case PE_ArrowDown: {
                    a.setPoints(7, d_arrow);
                    break;
                }
                case PE_ArrowLeft: {
                    a.setPoints(7, l_arrow);
                    break;
                }
                case PE_ArrowRight: {
                    a.setPoints(7, r_arrow);
                    break;
                }
                default: {
                    if (flags & Style_Up) {
                        a.setPoints(7, u_arrow);
                    } else {
                        a.setPoints(7, d_arrow);
                    }
                }
            }

            p->save();

            if (flags & Style_Down) {
                p->translate(pixelMetric(PM_ButtonShiftHorizontal),
                                pixelMetric(PM_ButtonShiftVertical));
            }

            a.translate((r.x()+r.width()/2), (r.y()+r.height()/2));
            // extra-pixel-shift, correcting some visual tics...
            switch(pe) {
                case PE_ArrowLeft:
                case PE_ArrowRight:
                    a.translate(0, -1);
                    break;
                case PE_SpinWidgetUp:
                case PE_SpinWidgetDown:
                    a.translate(+1, 0);
                    break;
                default:
                    a.translate(0, 0);
            }

            if (p->pen() == QPen::NoPen) {
                if (flags & Style_Enabled) {
                    p->setPen(cg.buttonText());
                } else {
                    p->setPen(cg.highlightedText());
                }
            }
            p->drawLineSegments(a, 0, 3);
            p->drawPoint(a[6]);

            p->restore();
            break;
        }

        default: {
            return KStyle::drawPrimitive(pe, p, r, cg, flags, opt);
        }
    }
}


void PlastikStyle::drawControl(ControlElement element,
                              QPainter *p,
                              const QWidget *widget,
                              const QRect &r,
                              const QColorGroup &cg,
                              SFlags flags,
                              const QStyleOption& opt) const
{
    const bool enabled = (flags & Style_Enabled);
    const bool hasFocus = (flags & Style_HasFocus);

    switch (element) {

    // PROGRESSBAR
    // -----------
        case CE_ProgressBarGroove: {
            const QColor content = enabled?cg.base():cg.background();
            const QColor button = enabled?cg.button():cg.background();
            renderContour(p, r, cg.background(), button.dark(150) );
            p->setPen(content.dark(105) );
            p->drawLine(r.left()+2, r.top()+1, r.right()-2, r.top()+1 );
            p->drawLine(r.left()+1, r.top()+2, r.left()+1, r.bottom()-2);
            p->setPen(content.light(105) );
            p->drawLine(r.left()+2, r.bottom()-1, r.right()-2, r.bottom()-1 );
            p->drawLine(r.right()-1, r.top()+2, r.right()-1, r.bottom()-2);
            break;
        }

        case CE_ProgressBarContents: {
            p->save();

            const QProgressBar *pb = dynamic_cast<const QProgressBar*>(widget);

            const QColor bg = enabled?cg.base():cg.background(); // background
            const QColor fg = enabled?cg.highlight():cg.background().dark(110); // foreground

            if (pb->totalSteps()) {
                double percent = static_cast<double>(pb->progress()) / static_cast<double>(pb->totalSteps());

                int w = static_cast<int>(r.width() * percent);
                // renderContour/renderSourface handle small sizes not very well, so set a minimal
                // progressbar width...
                if(w<3) w = 3;
                int w2 = r.width()-(r.width()-w);

                QRect Rempty(reverseLayout?r.left():r.left()+w, r.top(), r.width()-w, r.height() );
                QRect Rcontour(reverseLayout?r.right()-w2+1:r.left(), r.top(), w2, r.height() );
                QRect Rsourface(Rcontour.left()+1, Rcontour.top()+1, w2-2, Rcontour.height()-2);

                p->fillRect(Rempty, bg);

                renderContour(p, Rcontour,
                              bg, fg.dark(150),
                              reverseLayout ? Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Round_UpperLeft|Round_BottomLeft
                              : Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Round_UpperRight|Round_BottomRight);

                QRegion mask(Rsourface);
                if(reverseLayout) {
                    mask -= QRegion(Rsourface.left(), Rsourface.top(), 1, 1);
                    mask -= QRegion(Rsourface.left(), Rsourface.bottom(), 1, 1);
                }
                p->setClipRegion(mask);
                p->setClipping(true);
                int counter = 0;
                while((counter*10) < Rsourface.width()) {
                    counter++;
                    renderSourface(p, QRect(Rsourface.right()-counter*10, r.top()+1, 11, r.height()-2),
                                   (counter%2 == 0)?fg.light(105):fg, (counter%2 == 0)?fg:fg.light(105),
                                   cg.highlight(), 5,
                                   reverseLayout ? Draw_Right|Draw_Left|Draw_Top|Draw_Bottom|
                                           Round_UpperLeft|Round_BottomLeft|Is_Horizontal
                                   : Draw_Right|Draw_Left|Draw_Top|Draw_Bottom|
                                           Round_UpperRight|Round_BottomRight|Is_Horizontal);
                }
                p->setClipping(false);

                p->restore();
            }
            break;
        }

    // RADIOBUTTONS
    // ------------
        case CE_RadioButton: {
            drawPrimitive(PE_ExclusiveIndicator, p, r, cg, flags);

            const QColor contentColor = enabled?cg.base():cg.background();
            QColor checkmarkColor = enabled?cg.highlight():cg.background();
            if(flags & Style_Down) {
                checkmarkColor = alphaBlendColors(contentColor, checkmarkColor, 150);
            }

            p->save();

            if (flags & Style_On || flags & Style_Down) {
                int x = r.center().x() - 4, y = r.center().y() - 4;
                QBitmap bmp;
                bmp = QBitmap(CHECKMARKSIZE, CHECKMARKSIZE, radiomark_dark_bits, true);
                bmp.setMask(bmp);
                p->setPen(alphaBlendColors(contentColor, checkmarkColor.dark(150), 50) );
                p->drawPixmap(x, y, bmp);
                bmp = QBitmap(CHECKMARKSIZE, CHECKMARKSIZE, radiomark_light_bits, true);
                bmp.setMask(bmp);
                p->setPen(alphaBlendColors(contentColor, checkmarkColor.dark(125), 50) );
                p->drawPixmap(x, y, bmp);
                bmp = QBitmap(CHECKMARKSIZE, CHECKMARKSIZE, radiomark_aa_bits, true);
                bmp.setMask(bmp);
                p->setPen(alphaBlendColors(contentColor, checkmarkColor.dark(150), 150) );
                p->drawPixmap(x, y, bmp);
            }

            p->restore();

            break;
        }

    // TABS
    // ----
        case CE_TabBarTab: {
            const QTabBar * tb = (const QTabBar *) widget;
            QTabBar::Shape tbs = tb->shape();
            bool selected = false;
            if (flags & Style_Selected) selected = true;
            TabPosition pos;
            if (tb->count() == 1) {
                pos = Single;
            } else if ((tb->indexOf(opt.tab()->identifier()) == 0)) {
                pos = First;
            } else if (tb->indexOf(opt.tab()->identifier()) == tb->count() - 1) {
                pos = Last;
            } else {
                pos = Middle;
            }

            switch (tbs) {
// who really uses triangular tabs might be happier with the original ;)
// case QTabBar::TriangularAbove:
                case QTabBar::RoundedAbove:
                    renderTab(p, r, cg, (flags & Style_MouseOver), selected, false, pos);
                    break;
// case QTabBar::TriangularBelow:
                case QTabBar::RoundedBelow:
                    renderTab(p, r, cg, (flags & Style_MouseOver), selected, true, pos);
                    break;
                    default:
                            KStyle::drawControl(element, p, widget, r, cg, flags, opt);
            }

            break;
        }

        case CE_PushButton: {
            QPushButton *button = (QPushButton *)widget;

            if (widget == hoverWidget)
                flags |= Style_MouseOver;

            drawPrimitive(PE_ButtonBevel, p, r, cg, flags);

            if(hasFocus) {
                drawPrimitive(PE_FocusRect, p,
                        QRect(r.x()+3, r.y()+3, r.width()-6, r.height()-6), cg);
            }


            if (button->isDefault() ) {
                drawPrimitive(PE_ButtonDefault, p, r, cg, flags);
            }

            break;
        }

        case CE_PushButtonLabel: {
            const QPushButton *pb = (const QPushButton *)widget;
            const bool enabled = flags & Style_Enabled;
            const int text_flags = AlignVCenter | AlignHCenter | ShowPrefix |
                                    DontClip | SingleLine;
            int dx = 0;

            if (flags & Style_Down) {
                p->translate(pixelMetric(PM_ButtonShiftHorizontal),
                                pixelMetric(PM_ButtonShiftVertical));
            }

            if (pb->text() && flags & Style_ButtonDefault) {
                p->setFont(QFont(p->font().family(), p->font().pointSize(), 75));
            }

            if (pb->iconSet() && !pb->iconSet()->isNull()) {
                QIconSet::Mode mode = enabled ? QIconSet::Normal : QIconSet::Disabled;
                QPixmap pixmap = pb->iconSet()->pixmap(QIconSet::Small, mode);
                QRect pr(0, 0, pixmap.width(), pixmap.height());

                pr.moveCenter(r.center());

                if (pb->text()) {
                    pr.moveCenter(r.center());
                    dx = r.height() - pixmap.height();
                    pr.moveLeft(dx);
                }

                dx = static_cast<int>(pixmap.width() * 0.5);

                p->drawPixmap(pr.topLeft(), pixmap);
            } else if (pb->pixmap() && !pb->text()) {
                QRect pr(0, 0, pb->pixmap()->width(), pb->pixmap()->height());
                pr.moveCenter(r.center());
                p->drawPixmap(pr.topLeft(), *pb->pixmap());
            }

            if (pb->text()) {
                p->setPen(enabled ? cg.foreground() : pb->palette().disabled().buttonText());
                if (pb->iconSet() && !pb->iconSet()->isNull()) {
                    QRect tr(p->boundingRect(r, text_flags, pb->text()));
                    tr.moveCenter(r.center());
                    tr.moveBy(dx, 0);
                    p->drawText(tr, text_flags, pb->text());
                } else {
                    p->drawText(r, text_flags, pb->text());
                }
            }

            break;
        }

    // MENUBAR ITEM (sunken panel on mouse over)
    // -----------------------------------------
        case CE_MenuBarItem: {
            QMenuItem *mi = opt.menuItem();
            bool active  = flags & Style_Active;
            bool focused = flags & Style_HasFocus;
            bool down = flags & Style_Down;
            const int text_flags =
                AlignVCenter | AlignHCenter | ShowPrefix | DontClip | SingleLine;

            p->fillRect(r, cg.background());

            if (active && focused) {
                if (down) {
                    drawPrimitive(PE_ButtonTool, p, r, cg, flags|Style_Down, opt);
                } else {
                    drawPrimitive(PE_ButtonTool, p, r, cg, flags, opt);
                }
            }

            p->setPen(cg.foreground() );
            p->drawText(r, text_flags, mi->text());
            break;
        }

    // POPUPMENU ITEM (highlighted on mouseover)
    // ------------------------------------------
        case CE_PopupMenuItem: {
            const QPopupMenu *popupmenu = static_cast< const QPopupMenu * >( widget );
            QMenuItem *mi = opt.menuItem();

            if ( !mi )
            {
                // Don't leave blank holes if we set NoBackground for the QPopupMenu.
                // This only happens when the popupMenu spans more than one column.
                if (! ( widget->erasePixmap() && !widget->erasePixmap()->isNull() ) )
                    p->fillRect( r, cg.background().light( 105 ) );

                break;
            }
            int  tab        = opt.tabWidth();
            int  checkcol   = opt.maxIconWidth();
            bool enabled    = mi->isEnabled();
            bool checkable  = popupmenu->isCheckable();
            bool active     = flags & Style_Active;
            bool etchtext   = styleHint( SH_EtchDisabledText );
            bool reverse    = QApplication::reverseLayout();
            if ( checkable )
                checkcol = QMAX( checkcol, 20 );

            // Draw the menu item background
            if (active) {
                if (enabled) {
                    renderSourface(p, r, cg.background(), cg.highlight(), cg.highlight(),
                            10, Draw_Top|Draw_Bottom|Is_Horizontal);
                }
                else {
                    if ( widget->erasePixmap() && !widget->erasePixmap()->isNull() )
                        p->drawPixmap( r.topLeft(), *widget->erasePixmap(), r );
                    else p->fillRect( r, cg.background().light(105) );
                    p->drawWinFocusRect( r );
                }
            }
            // Draw the transparency pixmap
            else if ( widget->erasePixmap() && !widget->erasePixmap()->isNull() )
                p->drawPixmap( r.topLeft(), *widget->erasePixmap(), r );
            // Draw a solid background
            else
                p->fillRect( r, cg.background().light( 105 ) );
            // Are we a menu item separator?
            if ( mi->isSeparator() )
            {
                p->setPen( cg.mid() );
                p->drawLine( r.x()+5, r.y() + 1, r.width()-10, r.y() + 1 );
                p->setPen( cg.light() );
                p->drawLine( r.x()+5, r.y() + 2, r.width()-10 , r.y() + 2 );
                break;
            }

            QRect cr = visualRect( QRect( r.x() + 2, r.y() + 2, checkcol - 1, r.height() - 4 ), r );
            // Do we have an icon?
            if ( mi->iconSet() )
            {
                QIconSet::Mode mode;

                // Select the correct icon from the iconset
                if (active)
                    mode = enabled?QIconSet::Active:QIconSet::Disabled;
                else
                    mode = enabled?QIconSet::Normal:QIconSet::Disabled;

                // Do we have an icon and are checked at the same time?
                // Then draw a "pressed" background behind the icon
                if ( checkable && /*!active &&*/ mi->isChecked() )
                    qDrawShadePanel( p, cr.x(), cr.y(), cr.width(), cr.height(),
                                        cg, true, 1, &cg.brush(QColorGroup::Midlight) );
                // Draw the icon
                QPixmap pixmap = mi->iconSet()->pixmap(QIconSet::Small, mode);
                QRect pmr( 0, 0, pixmap.width(), pixmap.height() );
                pmr.moveCenter( cr.center() );
                p->drawPixmap( pmr.topLeft(), pixmap );
            }

            // Are we checked? (This time without an icon)
            else if ( checkable && mi->isChecked() )
            {
                // We only have to draw the background if the menu item is inactive -
                // if it's active the "pressed" background is already drawn
            // if ( ! active )
                    qDrawShadePanel( p, cr.x(), cr.y(), cr.width(), cr.height(), cg, true, 1,
                                        &cg.brush(QColorGroup::Midlight) );

                // Draw the checkmark
                SFlags cflags = Style_Default;
                cflags |= active ? Style_Enabled : Style_On;

                drawPrimitive( PE_CheckMark, p, cr, cg, cflags );
            }

            // Time to draw the menu item label...
            int xm = 2 + checkcol + 2; // X position margin

            int xp = reverse ? // X position
                    r.x() + tab + 1 + 2 + 1 - 1 :
                    r.x() + xm;

            int offset = reverse ? -1 : 1; // Shadow offset for etched text

            // Label width (minus the width of the accelerator portion)
            int tw = r.width() - xm - tab - 5 - 2 * 3 - 1 + 1;

            // Set the color for enabled and disabled text
            // (used for both active and inactive menu items)
            p->setPen( enabled ? cg.buttonText() : cg.mid() );

            // This color will be used instead of the above if the menu item
            // is active and disabled at the same time. (etched text)
            QColor discol = cg.mid();

            // Does the menu item draw it's own label?
            if ( mi->custom() ) {
                int m = 2;
                // Save the painter state in case the custom
                // paint method changes it in some way
                p->save();

                // Draw etched text if we're inactive and the menu item is disabled
                if ( etchtext && !enabled && !active ) {
                    p->setPen( cg.light() );
                    mi->custom()->paint( p, cg, active, enabled, xp+offset, r.y()+m+1, tw, r.height()-2*m );
                    p->setPen( discol );
                }
                mi->custom()->paint( p, cg, active, enabled, xp, r.y()+m, tw, r.height()-2*m );
                p->restore();
            }
            else {
                // The menu item doesn't draw it's own label
                QString s = mi->text();
                // Does the menu item have a text label?
                if ( !s.isNull() ) {
                    int t = s.find( '\t' );
                    int m = 2;
                    int text_flags = AlignVCenter | ShowPrefix | DontClip | SingleLine;
                    text_flags |= reverse ? AlignRight : AlignLeft;

                    //QColor draw = cg.text();
                    QColor draw = (active && enabled) ? cg.highlightedText () : cg.text();
                    p->setPen(draw);


                    // Does the menu item have a tabstop? (for the accelerator text)
                    if ( t >= 0 ) {
                        int tabx = reverse ? r.x() + 1 + 2 + 1 :
                            r.x() + r.width() - tab - 1 - 2 - 1;

                        // Draw the right part of the label (accelerator text)
                        if ( etchtext && !enabled ) {
                            // Draw etched text if we're inactive and the menu item is disabled
                            p->setPen( cg.light() );
                            p->drawText( tabx+offset, r.y()+m+1, tab, r.height()-2*m, text_flags, s.mid( t+1 ) );
                            p->setPen( discol );
                        }
                        p->drawText( tabx, r.y()+m, tab, r.height()-2*m, text_flags, s.mid( t+1 ) );
                        s = s.left( t );
                    }

                    // Draw the left part of the label (or the whole label
                    // if there's no accelerator)
                    if ( etchtext && !enabled ) {
                        // Etched text again for inactive disabled menu items...
                        p->setPen( cg.light() );
                        p->drawText( xp+offset, r.y()+m+1, tw, r.height()-2*m, text_flags, s, t );
                        p->setPen( discol );
                    }


                    p->drawText( xp, r.y()+m, tw, r.height()-2*m, text_flags, s, t );

                    p->setPen(cg.text());

                }

                // The menu item doesn't have a text label
                // Check if it has a pixmap instead
                else if ( mi->pixmap() ) {
                    QPixmap *pixmap = mi->pixmap();

                    // Draw the pixmap
                    if ( pixmap->depth() == 1 )
                        p->setBackgroundMode( OpaqueMode );

                    int diffw = ( ( r.width() - pixmap->width() ) / 2 )
                                    + ( ( r.width() - pixmap->width() ) % 2 );
                    p->drawPixmap( r.x()+diffw, r.y()+1, *pixmap );

                    if ( pixmap->depth() == 1 )
                        p->setBackgroundMode( TransparentMode );
                }
            }

            // Does the menu item have a submenu?
            if ( mi->popup() ) {
                PrimitiveElement arrow = reverse ? PE_ArrowLeft : PE_ArrowRight;
                int dim = pixelMetric(PM_MenuButtonIndicator) - 1;
                QRect vr = visualRect( QRect( r.x() + r.width() - 5 - 1 - dim,
                            r.y() + r.height() / 2 - dim / 2, dim, dim), r );

                // Draw an arrow at the far end of the menu item
                if ( active ) {
                    if ( enabled )
                        discol = cg.buttonText();

                    QColorGroup g2( discol, cg.highlight(), white, white,
                                    enabled ? white : discol, discol, white );

                    drawPrimitive( arrow, p, vr, g2, Style_Enabled );
                } else
                    drawPrimitive( arrow, p, vr, cg,
                            enabled ? Style_Enabled : Style_Default );
            }
            break;
        }

    // Menu and dockwindow empty space
    //
        case CE_DockWindowEmptyArea:
        case CE_MenuBarEmptyArea:
            p->fillRect(r, cg.background());
            break;

        default:
          KStyle::drawControl(element, p, widget, r, cg, flags, opt);
    }
}

void PlastikStyle::drawControlMask(ControlElement element,
                                  QPainter *p,
                                  const QWidget *w,
                                  const QRect &r,
                                  const QStyleOption &opt) const
{
    switch (element) {
        case CE_PushButton: {
                    p->fillRect (r, color0);
                    renderMask(p, r, color1,
                            Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight);
                    break;
        }

        default: {
            KStyle::drawControlMask (element, p, w, r, opt);
        }
    }
}

void PlastikStyle::drawComplexControlMask(ComplexControl c,
                                         QPainter *p,
                                         const QWidget *w,
                                         const QRect &r,
                                         const QStyleOption &o) const
{
    switch (c) {
        case CC_SpinWidget:
        case CC_ListView:
        case CC_ComboBox: {
                p->fillRect (r, color0);
                renderMask(p, r, color1,
                        Round_UpperLeft|Round_UpperRight|Round_BottomLeft|Round_BottomRight);
            break;
        }
        default: {
            KStyle::drawComplexControlMask (c, p, w, r, o);
        }
    }
}

void PlastikStyle::drawComplexControl(ComplexControl control,
                                     QPainter *p,
                                     const QWidget *widget,
                                     const QRect &r,
                                     const QColorGroup &cg,
                                     SFlags flags,
                                     SCFlags controls,
                                     SCFlags active,
                                     const QStyleOption& opt) const
{
    const bool enabled = (flags & Style_Enabled);

    switch(control) {
    // COMBOBOX
    // --------
        case CC_ComboBox: {
            static const unsigned int handleWidth = 15;

            const QComboBox *cb = dynamic_cast<const QComboBox *>(widget);
            bool editable;
            // at the moment cb is only needed to check if the combo box is editable or not.
            // if cb doesn't exist, just assume false and the app (gideon! ;) ) at least doesn't crash.
            if(!cb) {
                editable = false;
            } else {
                editable = cb->editable();
            }

            const QColor buttonColor = enabled?cg.button():cg.background();
            const QColor inputColor = enabled?(editable?cg.base():cg.button())
                                              :cg.background();

            p->save();

            renderContour(p, r, cg.background(), buttonColor.dark(150));
            //extend the contour: between input and handler...
            p->setPen(alphaBlendColors(cg.background(), buttonColor.dark(150), 50) );
            if(reverseLayout) {
                p->drawLine(r.left()+1+handleWidth, r.top()+1, r.left()+1+handleWidth, r.bottom()-1);
            } else {
                p->drawLine(r.right()-handleWidth-1, r.top()+1, r.right()-handleWidth-1, r.bottom()-1);
            }

            const QRect RbuttonSourface(reverseLayout?r.left()+1:r.right()-handleWidth, r.top()+1,
                                        handleWidth, r.height()-2);
            const QRect RcontentSourface(reverseLayout?r.left()+1+handleWidth+1:r.left()+1, r.top()+1,
                                         r.width()-handleWidth-3, r.height()-2);

            // handler

            uint sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Is_Horizontal;
            if(reverseLayout) {
                sourfaceFlags |= Round_UpperLeft|Round_BottomLeft;
            } else {
                sourfaceFlags |= Round_UpperRight|Round_BottomRight;
            }

            if (widget == hoverWidget) {
                sourfaceFlags |= Is_Highlight;
                if(editable) sourfaceFlags |= Highlight_Left|Highlight_Right;
                sourfaceFlags |= Highlight_Top|Highlight_Bottom;
            }
            renderSourface(p, RbuttonSourface,
                           cg.background(), buttonColor, cg.highlight(), enabled?10:2,
                           sourfaceFlags);

            if(!editable) {
                sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Is_Horizontal;
                if(reverseLayout) {
                    sourfaceFlags |= Round_UpperRight|Round_BottomRight;
                } else {
                    sourfaceFlags |= Round_UpperLeft|Round_BottomLeft;
                }

                if (widget == hoverWidget) {
                    sourfaceFlags |= Is_Highlight;
                    sourfaceFlags |= Highlight_Top|Highlight_Bottom;
                }
                renderSourface(p, RcontentSourface,
                               cg.background(), buttonColor, cg.highlight(), enabled?10:2,
                               sourfaceFlags);
            } else {
                // thin frame around the input area
                p->setPen(inputColor.dark(130) );
                p->drawLine(RcontentSourface.x(), reverseLayout?RcontentSourface.y():RcontentSourface.y()+1,
                        RcontentSourface.x(), reverseLayout?RcontentSourface.bottom():RcontentSourface.bottom()-1);
                p->drawLine(RcontentSourface.x()+1, RcontentSourface.y(),
                        reverseLayout?RcontentSourface.right()-1:RcontentSourface.right(), RcontentSourface.y() );
                p->setPen(inputColor.light(130) );
                p->drawLine(reverseLayout?RcontentSourface.x():RcontentSourface.x()+1, RcontentSourface.bottom(),
                        reverseLayout?RcontentSourface.right()-1:RcontentSourface.right(), RcontentSourface.bottom() );
                p->drawLine(RcontentSourface.right(), RcontentSourface.top()+1,
                        RcontentSourface.right(), RcontentSourface.bottom()-1 );

                // input area
                p->fillRect(RcontentSourface.x()+1, RcontentSourface.y()+1,
                        RcontentSourface.width()-2, RcontentSourface.height()-2, inputColor );
            }

            p->setPen(cg.foreground());
            drawPrimitive(PE_SpinWidgetDown, p, RbuttonSourface, cg, Style_Default|Style_Enabled|Style_Raised);

            p->restore();
            break;
        }

    // TOOLBUTTON
    // ----------
        case CC_ToolButton: {
            const QToolButton *tb = (const QToolButton *) widget;

            QRect button, menuarea;
            button   = querySubControlMetrics(control, widget, SC_ToolButton, opt);
            menuarea = querySubControlMetrics(control, widget, SC_ToolButtonMenu, opt);

            SFlags bflags = flags,
                    mflags = flags;

            if (kornMode) {
                drawPrimitive(PE_ButtonTool, p, button, cg, bflags, opt);
                break;
            } else {
                // don't want to have the buttoncolor as the background...
                p->fillRect(r, cg.background());
                bflags &= ~Style_MouseOver;
            }

            if (active & SC_ToolButton)
                bflags |= Style_Down;

            if (active & SC_ToolButtonMenu)
                mflags |= Style_Down;

            if (controls & SC_ToolButton) {
            // If we're pressed, on, or raised...
                if (bflags & (Style_Down | Style_On | Style_Raised)) {
                    drawPrimitive(PE_ButtonTool, p, button, cg, bflags, opt);
                } else if (tb->parentWidget() &&
                            tb->parentWidget()->backgroundPixmap() &&
                            !tb->parentWidget()->backgroundPixmap()->isNull()) {
                    QPixmap pixmap = *(tb->parentWidget()->backgroundPixmap());
                    p->drawTiledPixmap( r, pixmap, tb->pos() );
                }
            }

            // Draw a toolbutton menu indicator if required
            if (controls & SC_ToolButtonMenu) {
                if (mflags & (Style_Down | Style_On | Style_Raised)) {
                    drawPrimitive(PE_ButtonDropDown, p, menuarea, cg, mflags, opt);
                }
                drawPrimitive(PE_ArrowDown, p, menuarea, cg, mflags, opt);
            }

            if (tb->hasFocus() && !tb->focusProxy()) {
                QRect fr = tb->rect();
                fr.addCoords(2, 2, -2, -2);
                drawPrimitive(PE_FocusRect, p, fr, cg);
            }

            break;
        }

    // SPINWIDGETS
    // -----------
        case CC_SpinWidget: {
            static const unsigned int handleWidth = 15;

            const QSpinWidget *sw = dynamic_cast<const QSpinWidget *>(widget);
            SFlags flags;
            PrimitiveElement pe;

            const QColor buttonColor = enabled?cg.button():cg.background();
            const QColor inputColor = enabled?cg.base():cg.background();

            p->save();

            // contour
            const bool heightDividable = ((r.height()%2) == 0);
            renderContour(p, querySubControlMetrics(control, widget, SC_SpinWidgetFrame),
                          cg.background(), enabled?buttonColor.dark(150):cg.background().dark(150) );
            p->setPen(alphaBlendColors(cg.background(), enabled?buttonColor.dark(150):cg.background().dark(150), 50) );
            p->drawLine(reverseLayout?r.left()+1+handleWidth:r.right()-handleWidth-1, r.top()+1,
                    reverseLayout?r.left()+1+handleWidth:r.right()-handleWidth-1, r.bottom()-1);
            p->drawLine(reverseLayout?r.left()+1:r.right()-handleWidth, r.top()+1+(r.height()-2)/2,
                    reverseLayout?r.left()+handleWidth:r.right()-1, r.top()+1+(r.height()-2)/2);
            if(heightDividable)
                p->drawLine(reverseLayout?r.left()+1:r.right()-handleWidth, r.top()+1+(r.height()-2)/2-1,
                        reverseLayout?r.left()+handleWidth:r.right()-1, r.top()+1+(r.height()-2)/2-1);

            // sourface
//             // doesn't work properly with reverse layout.
//             QRect upRect = querySubControlMetrics(control, widget, SC_SpinWidgetUp);
//             QRect downRect = querySubControlMetrics(control, widget, SC_SpinWidgetDown);
            QRect upRect = QRect(reverseLayout?r.left()+1:r.right()-handleWidth, r.top()+1,
                    handleWidth, (r.height()-2)/2);
            QRect downRect = QRect(reverseLayout?r.left()+1:r.right()-handleWidth,
                    heightDividable?r.top()+1+((r.height()-2)/2):r.top()+1+((r.height()-2)/2)+1,
                    handleWidth, ((r.height()-2)/2) );
            if(heightDividable) {
                upRect = QRect(upRect.left(), upRect.top(), upRect.width(), upRect.height()-1 );
                downRect = QRect(downRect.left(), downRect.top()+1, downRect.width(), downRect.height()-1 );
            }

            uint sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Is_Horizontal;
            if(reverseLayout) {
                sourfaceFlags |= Round_UpperLeft;
            } else {
                sourfaceFlags |= Round_UpperRight;
            }
            if (widget == hoverWidget) {
                sourfaceFlags |= Is_Highlight;
                sourfaceFlags |= Highlight_Top|Highlight_Left|Highlight_Right;
            }
            if (active==SC_SpinWidgetUp) sourfaceFlags|=Is_Sunken;
            if(!enabled) sourfaceFlags|=Is_Disabled;
            renderSourface(p, upRect, cg.background(), buttonColor, cg.highlight(),
                           5, sourfaceFlags);
            sourfaceFlags = Draw_Left|Draw_Right|Draw_Top|Draw_Bottom|Is_Horizontal;
            if(reverseLayout) {
                sourfaceFlags |= Round_BottomLeft;
            } else {
                sourfaceFlags |= Round_BottomRight;
            }
            if (widget == hoverWidget) {
                sourfaceFlags |= Is_Highlight;
                sourfaceFlags |= Highlight_Bottom|Highlight_Left|Highlight_Right;
            }
            if (active==SC_SpinWidgetDown) sourfaceFlags|=Is_Sunken;
            if(!enabled) sourfaceFlags|=Is_Disabled;
            renderSourface(p, downRect, cg.background(), buttonColor, cg.highlight(),
                           5, sourfaceFlags);

            // icons...
            flags = Style_Default | Style_Enabled;
            if (active == SC_SpinWidgetUp) {
                flags |= Style_On;
                flags |= Style_Sunken;
            } else
                flags |= Style_Raised;
            if (sw->buttonSymbols() == QSpinWidget::PlusMinus)
                pe = PE_SpinWidgetPlus;
            else
                pe = PE_SpinWidgetUp;
            p->setPen(cg.foreground());
            drawPrimitive(pe, p, upRect, cg, flags);

            flags = Style_Default | Style_Enabled;
            if (active == SC_SpinWidgetDown) {
                flags |= Style_On;
                flags |= Style_Sunken;
            } else
                flags |= Style_Raised;
            if (sw->buttonSymbols() == QSpinWidget::PlusMinus)
                pe = PE_SpinWidgetMinus;
            else
                pe = PE_SpinWidgetDown;
            p->setPen(cg.foreground());
            drawPrimitive(pe, p, downRect, cg, flags);

            // thin frame around the input area
            const QRect Rcontent = QRect(reverseLayout?r.left()+1+handleWidth+1:r.left()+1, r.top()+1,
                    r.width()-1-2-handleWidth, r.height()-2);
            p->setPen(inputColor.dark(130) );
            p->drawLine(Rcontent.left(), reverseLayout?Rcontent.top():Rcontent.top()+1,
                    Rcontent.left(), reverseLayout?Rcontent.bottom():Rcontent.bottom()-1 );
            p->drawLine(Rcontent.left()+1, Rcontent.top(),
                    reverseLayout?Rcontent.right()-1:Rcontent.right(), Rcontent.top() );
            p->setPen(inputColor.light(130) );
            p->drawLine(Rcontent.left()+1, Rcontent.bottom(), Rcontent.right()-1, Rcontent.bottom() );
            p->drawLine(Rcontent.right(), Rcontent.top()+1,
                    Rcontent.right(), reverseLayout?Rcontent.bottom()-1:Rcontent.bottom() );

            p->restore();

            break;
        }

        default:
            KStyle::drawComplexControl(control, p, widget,
                                        r, cg, flags, controls,
                                        active, opt);
            break;
    }
}


QRect PlastikStyle::subRect(SubRect r, const QWidget *widget) const
{
    switch (r) {
        case SR_ComboBoxFocusRect: {
            return querySubControlMetrics( CC_ComboBox, widget, SC_ComboBoxEditField );
        }

        // Don't use KStyles progressbar subrect
        // TODO:
        case SR_ProgressBarGroove: {
            return QRect(widget->rect());
        }
        case SR_ProgressBarContents:
        case SR_ProgressBarLabel: {
            QRect rw = widget->rect();
            return QRect(rw.left()+2, rw.top()+2, rw.width()-4, rw.height()-4 );
        }

        default: {
            return KStyle::subRect(r, widget);
        }
    }
}

QRect PlastikStyle::querySubControlMetrics(ComplexControl control,
                                          const QWidget *widget,
                                          SubControl subcontrol,
                                          const QStyleOption &opt) const
{
    if (!widget) {
        return QRect();
    }

    QRect r(widget->rect());
    switch (control) {
        case CC_ComboBox: {
            switch (subcontrol) {
                case SC_ComboBoxEditField: {
                    // TODO: is the handler width in pixelmetric?
                    return QRect(r.left()+2, r.top()+2, r.width()-4-15-1, r.height()-4);
                }
                default: {
                    return KStyle::querySubControlMetrics(control, widget, subcontrol, opt);
                }
            }
            break;
        }
        case CC_SpinWidget: {
            const int fw = 2; // Frame width...

            const bool heightDividable = ((r.height()%2) == 0);

            QSize bs;
            if(heightDividable) {
                bs.setHeight(QMAX(8, (r.height()-2)/2));
            } else {
                bs.setHeight(QMAX(8, (r.height()-2-1)/2));
            }
            bs.setWidth(15);

            const int buttonsLeft = /*reverseLayout?r.left()+1:*/r.right()-bs.width();

            switch (subcontrol) {
                case SC_SpinWidgetUp: {
                    return QRect(buttonsLeft, r.top()+1, bs.width(), bs.height() );
                }
                case SC_SpinWidgetDown: {
                    if(heightDividable) {
                        return QRect(buttonsLeft, r.top()+1+bs.height(),
                                bs.width(), r.height()-(bs.height()+2) );
                    } else {
                        return QRect(buttonsLeft, r.top()+1+bs.height()+1,
                                bs.width(), r.height()-(bs.height()+2+1) );
                    }
                }
                case SC_SpinWidgetFrame: {
                    return QRect(r.left(), r.top(), r.width(), r.height() );
                }
                case SC_SpinWidgetEditField: {
                    return QRect(r.left()+fw, r.top()+fw,
                            r.width()-(bs.width()+1+2*fw), r.height()-2*fw);
                }
                case SC_SpinWidgetButtonField: {
                    return QRect(buttonsLeft, r.top()+1, bs.width(), r.height()-2);
                }
                default: {
                    return KStyle::querySubControlMetrics(control, widget, subcontrol, opt);
                }
            }
            break;
        }
        default: {
            return KStyle::querySubControlMetrics(control, widget, subcontrol, opt);
        }
    }
}

int PlastikStyle::pixelMetric(PixelMetric m, const QWidget *widget) const
{
    switch(m) {
    // TABS
    // ----
        case PM_TabBarTabVSpace: {
            const QTabBar * tb = (const QTabBar *) widget;
            if (tb->shape() == QTabBar::RoundedAbove ||
                tb->shape() == QTabBar::RoundedBelow)
                return 12;
            else
                return 0;
        }

        case PM_TabBarTabOverlap: {
            const QTabBar* tb = (const QTabBar*)widget;
            if (tb->shape() != QTabBar::RoundedAbove &&
                tb->shape() != QTabBar::RoundedBelow)
                return 3;                   // Leave standard size alone
            else
                return 1;                   // Change size for our tabs only
        }

    // SCROLL BAR
        case PM_ScrollBarSliderMin: {
            return 20;
        }
        case PM_ScrollBarExtent: {
            return 16;
        }

    // SPLITTERS
    // ---------
        case PM_SplitterWidth: {
            return 5;
        }

    // PROGRESSBARS
    // ------------
        case PM_ProgressBarChunkWidth:
            return 10;

    // SLIDER
    // ------
        case PM_SliderLength:
            return 11;

    // MENU INDICATOR
    // --------------
        case PM_MenuButtonIndicator:
            return 8;

    // CHECKBOXES / RADIO BUTTONS
    // --------------------------
        case PM_ExclusiveIndicatorWidth:    // Radiobutton size
        case PM_ExclusiveIndicatorHeight:   // 13x13
        case PM_IndicatorWidth:             // Checkbox size
        case PM_IndicatorHeight:            // 13x13
            return 13;

    // FRAMES
    // ------
        case PM_SpinBoxFrameWidth:
            return 1;

        case PM_MenuBarFrameWidth:
            return 0;

        case PM_DefaultFrameWidth: {
            return 2;
        }

        case PM_ButtonDefaultIndicator: {
            return 0;
        }

        case PM_ButtonMargin: {
            return 2;
        }

        case PM_ButtonShiftVertical:
        case PM_ButtonShiftHorizontal: {
            return 1;
        }

        default:
            return KStyle::pixelMetric(m, widget);
    }
}


QSize PlastikStyle::sizeFromContents(ContentsType t,
                                    const QWidget *widget,
                                    const QSize &s,
                                    const QStyleOption &opt) const
{
    switch (t) {
        case CT_MenuBar: {
            const QMenuBar *mb = (const QMenuBar *)widget;
            int w, h;
            {
                h = QFontInfo(mb->font()).pixelSize() + 4;
                w = 10;
            }

            return QSize(w, h);

            break;
        }

        case CT_PopupMenuItem: {
            if (!widget || opt.isDefault())
                return s;

            const QPopupMenu *popup = dynamic_cast<const QPopupMenu *>(widget);
            QMenuItem *mi = opt.menuItem();
            int maxpmw = opt.maxIconWidth();
            int w = s.width(), h = s.height();
            bool checkable = popup->isCheckable();

            if (mi->custom()) {
                w = mi->custom()->sizeHint().width();
                h = mi->custom()->sizeHint().height();
                if (!mi->custom()->fullSpan() )
                    h += 4;
            } else if (mi->widget()) {
            // don't change the size in this case.
            } else if (mi->isSeparator()) {
                w = 20;
                h = 2;
            } else {
                if (mi->pixmap()) {
                    h = QMAX(h, mi->pixmap()->height() + 2);
                } else {
                    h = QMAX(h, 16 + 2 );
                    h = QMAX(h, popup->fontMetrics().height() + 4 );
                }

                if (mi->iconSet()) {
                    h = QMAX(h, mi->iconSet()->pixmap(QIconSet::Small, QIconSet::Normal).height() + 2);
                }
            }

            if (!mi->text().isNull() && (mi->text().find('\t') >= 0)) {
                w += 12;
            } else if (mi->popup()) {
                w += 24;
            }

            if (maxpmw) {
                w += maxpmw + 6;
            }
            if (checkable && maxpmw < 20) {
                w += 20 - maxpmw;
            }
            if (checkable || maxpmw > 0) {
                w += 12;
            }

            w += 20;

            return QSize(w, h);
        }

        case CT_PushButton:
        {
            const QPushButton* btn = static_cast<const QPushButton*>(widget);

            int w = s.width() + 2 * pixelMetric(PM_ButtonMargin, widget);
            int h = s.height() + 2 * pixelMetric(PM_ButtonMargin, widget);
            if ( btn->text().isEmpty() && s.width() < 32 ) return QSize(w, h);

            return QSize( w+25, h+5 );
        }

        default:
            return KStyle::sizeFromContents (t, widget, s, opt);
    }

    return KStyle::sizeFromContents (t, widget, s, opt);
}

// void PlastikStyle::paletteChanged()
// {
// 	QComboBox *combo  = 0L;
// 	QMenuBar *menuBar = 0L;
// 	QToolBar *toolBar = 0L;
//    QWidget  *widget  = 0L;
//
// 	QValueListIterator<QWidget*> it = m_widgets.begin();
// 	for ( ; it != m_widgets.end(); ++it ) {
// 		widget = *it;
// 		if ((combo = dynamic_cast<QComboBox*>( widget )))
// 			updatePalette( combo );
// 		else if ((toolBar = dynamic_cast<QToolBar*>( widget )))
// 			updatePalette( toolBar );
// 		else if ((menuBar = dynamic_cast<QMenuBar*>( widget )))
// 			updatePalette( menuBar );
// 	}
// }

// void PlastikStyle::updatePalette( QComboBox *combo )
// {
// 	QPalette pal = QApplication::palette();
// 	pal.setColor(QColorGroup::Dark,
// 	             pal.active().color(QColorGroup::Base));
// 	pal.setColor(QColorGroup::Midlight,
// 	             pal.active().color(QColorGroup::Background));
// 	combo->setPalette(pal);
// }
//
// void PlastikStyle::updatePalette( QToolBar *bar )
// {
// 	QPalette pal = QApplication::palette();
// 	pal.setColor(QColorGroup::Button,
// 	             pal.active().color(QColorGroup::Background));
// 	bar->setPalette(pal);
// }
//
// void PlastikStyle::updatePalette( QMenuBar *bar )
// {
// 	QPalette pal = QApplication::palette();
// 	pal.setColor(QColorGroup::Button,
// 	             pal.active().color(QColorGroup::Background));
// 	bar->setPalette(pal);
// }

// void PlastikStyle::slotDestroyed()
// {
// 	m_widgets.remove( (QWidget*) sender() );
// }

bool PlastikStyle::eventFilter(QObject *obj, QEvent *ev)
{
    if (KStyle::eventFilter(obj, ev) )
        return true;

    if (!obj->isWidgetType() ) return false;

    //Hover highlight...
    if (obj->inherits("QPushButton") || obj->inherits("QComboBox") || obj->inherits("QSpinWidget") ||
            obj->inherits("QSpinWidget") || obj->inherits("QCheckBox") ||
            obj->inherits("QRadioButton") )
    {
        if ((ev->type() == QEvent::Enter) && static_cast<QWidget*>(obj)->isEnabled())
        {
            QWidget* button = static_cast<QWidget*>(obj);
            hoverWidget = button;
            button->repaint(false);
        }
        else if ((ev->type() == QEvent::Leave) && (obj == hoverWidget) )
        {
            QWidget* button = static_cast<QWidget*>(obj);
            hoverWidget = 0;
            button->repaint(false);
        }
        return false;
    }

    return false;
}
